#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for Annotator properties."""
from typing import List, Dict

from pyspark.ml.param import Param, Params, TypeConverters


class HasBatchedAnnotate:
    batchSize = Param(Params._dummy(), "batchSize", "Size of every batch", TypeConverters.toInt)

    def setBatchSize(self, v):
        """Sets batch size.

        Parameters
        ----------
        v : int
            Batch size
        """
        return self._set(batchSize=v)

    def getBatchSize(self):
        """Gets current batch size.

        Returns
        -------
        int
            Current batch size
        """
        return self.getOrDefault(self.batchSize)


class HasCaseSensitiveProperties:
    caseSensitive = Param(Params._dummy(),
                          "caseSensitive",
                          "whether to ignore case in tokens for embeddings matching",
                          typeConverter=TypeConverters.toBoolean)

    def setCaseSensitive(self, value):
        """Sets whether to ignore case in tokens for embeddings matching.

        Parameters
        ----------
        value : bool
            Whether to ignore case in tokens for embeddings matching
        """
        return self._set(caseSensitive=value)

    def getCaseSensitive(self):
        """Gets whether to ignore case in tokens for embeddings matching.

        Returns
        -------
        bool
            Whether to ignore case in tokens for embeddings matching
        """
        return self.getOrDefault(self.caseSensitive)


class HasClsTokenProperties:
    useCLSToken = Param(Params._dummy(),
                        "useCLSToken",
                        "Whether to use CLS token for pooling (true) or attention-based average pooling (false)",
                        typeConverter=TypeConverters.toBoolean)

    def setUseCLSToken(self, value):
        """Sets whether to ignore case in tokens for embeddings matching.

        Parameters
        ----------
        value : bool
            Whether to use CLS token for pooling (true) or attention-based average pooling (false)
        """
        return self._set(useCLSToken=value)

    def getUseCLSToken(self):
        """Gets whether to use CLS token for pooling (true) or attention-based average pooling (false)

        Returns
        -------
        bool
            Whether to use CLS token for pooling (true) or attention-based average pooling (false)
        """
        return self.getOrDefault(self.useCLSToken)


class HasClassifierActivationProperties:
    activation = Param(Params._dummy(),
                       "activation",
                       "Whether to calculate logits via Softmax or Sigmoid. Default is Softmax",
                       typeConverter=TypeConverters.toString)

    multilabel = Param(Params._dummy(),
                       "multilabel",
                       "Whether to calculate logits via Multiclass(softmax) or Multilabel(sigmoid). Default is False i.e. Multiclass",
                       typeConverter=TypeConverters.toBoolean)

    threshold = Param(Params._dummy(),
                      "threshold",
                      "Choose the threshold to determine which logits are considered to be positive or negative",
                      typeConverter=TypeConverters.toFloat)

    def setActivation(self, value):
        """Sets whether to calculate logits via Softmax or Sigmoid. Default is Softmax

        Parameters
        ----------
        value : str
            Whether to calculate logits via Softmax or Sigmoid. Default is Softmax
        """
        return self._set(activation=value)

    def getActivation(self):
        """Gets whether to calculate logits via Softmax or Sigmoid. Default is Softmax

        Returns
        -------
        str
            Whether to calculate logits via Softmax or Sigmoid. Default is Softmax
        """
        return self.getOrDefault(self.activation)

    def setMultilabel(self, value):
        """Set whether or not the result should be multi-class (the sum of all probabilities is 1.0) or
         multi-label (each label has a probability between 0.0 to 1.0).
         Default is False i.e. multi-class

        Parameters
        ----------
        value : bool
            Whether or not the result should be multi-class (the sum of all probabilities is 1.0) or
            multi-label (each label has a probability between 0.0 to 1.0).
            Default is False i.e. multi-class
        """
        return self._set(multilabel=value)

    def getMultilabel(self):
        """Gets whether or not the result should be multi-class (the sum of all probabilities is 1.0) or
         multi-label (each label has a probability between 0.0 to 1.0).
         Default is False i.e. multi-class

        Parameters
        ----------
        value : bool
            Whether or not the result should be multi-class (the sum of all probabilities is 1.0) or
            multi-label (each label has a probability between 0.0 to 1.0).
            Default is False i.e. multi-class
        """
        return self.getOrDefault(self.multilabel)

    def setThreshold(self, value):
        """Set the threshold to determine which logits are considered to be positive or negative.
         (Default: `0.5`). The value should be between 0.0 and 1.0. Changing the threshold value
         will affect the resulting labels and can be used to adjust the balance between precision and
         recall in the classification process.

        Parameters
        ----------
        value : float
            The threshold to determine which logits are considered to be positive or negative.
            (Default: `0.5`). The value should be between 0.0 and 1.0. Changing the threshold value
            will affect the resulting labels and can be used to adjust the balance between precision and
            recall in the classification process.
        """
        return self._set(threshold=value)


class HasEmbeddingsProperties(Params):
    dimension = Param(Params._dummy(),
                      "dimension",
                      "Number of embedding dimensions",
                      typeConverter=TypeConverters.toInt)

    def setDimension(self, value):
        """Sets embeddings dimension.

        Parameters
        ----------
        value : int
            Embeddings dimension
        """
        return self._set(dimension=value)

    def getDimension(self):
        """Gets embeddings dimension."""
        return self.getOrDefault(self.dimension)


class HasEnableCachingProperties:
    enableCaching = Param(Params._dummy(),
                          "enableCaching",
                          "Whether to enable caching DataFrames or RDDs during the training",
                          typeConverter=TypeConverters.toBoolean)

    def setEnableCaching(self, value):
        """Sets whether to enable caching DataFrames or RDDs during the training

        Parameters
        ----------
        value : bool
            Whether to enable caching DataFrames or RDDs during the training
        """
        return self._set(enableCaching=value)

    def getEnableCaching(self):
        """Gets whether to enable caching DataFrames or RDDs during the training

        Returns
        -------
        bool
            Whether to enable caching DataFrames or RDDs during the training
        """
        return self.getOrDefault(self.enableCaching)


class HasBatchedAnnotateImage:
    batchSize = Param(Params._dummy(), "batchSize", "Size of every batch", TypeConverters.toInt)

    def setBatchSize(self, v):
        """Sets batch size.

        Parameters
        ----------
        v : int
            Batch size
        """
        return self._set(batchSize=v)

    def getBatchSize(self):
        """Gets current batch size.

        Returns
        -------
        int
            Current batch size
        """
        return self.getOrDefault(self.batchSize)


class HasImageFeatureProperties:
    doResize = Param(Params._dummy(), "doResize", "Whether to resize the input to a certain size",
                     TypeConverters.toBoolean)

    doNormalize = Param(Params._dummy(), "doNormalize",
                        "Whether to normalize the input with mean and standard deviation",
                        TypeConverters.toBoolean)

    featureExtractorType = Param(Params._dummy(), "featureExtractorType",
                                 "Name of model's architecture for feature extraction",
                                 TypeConverters.toString)

    imageMean = Param(Params._dummy(), "imageMean",
                      "The sequence of means for each channel, to be used when normalizing images",
                      TypeConverters.toListFloat)

    imageStd = Param(Params._dummy(), "imageStd",
                     "The sequence of standard deviations for each channel, to be used when normalizing images",
                     TypeConverters.toListFloat)

    resample = Param(Params._dummy(), "resample",
                     "An optional resampling filter. This can be one of PIL.Image.NEAREST, PIL.Image.BILINEAR or "
                     "PIL.Image.BICUBIC. Only has an effect if do_resize is set to True.",
                     TypeConverters.toInt)

    size = Param(Params._dummy(), "size",
                 "Resize the input to the given size. If a tuple is provided, it should be (width, height). If only "
                 "an integer is provided, then the input will be resized to (size, size). Only has an effect if "
                 "do_resize is set to True.",
                 TypeConverters.toInt)

    def setDoResize(self, value):
        """

        Parameters
        ----------
        value : Boolean
            Whether to resize the input to a certain size
        """
        return self._set(doResize=value)

    def setDoNormalize(self, value):
        """

        Parameters
        ----------
        value : Boolean
            Whether to normalize the input with mean and standard deviation
        """
        return self._set(doNormalize=value)

    def setFeatureExtractorType(self, value):
        """

        Parameters
        ----------
        value : str
            Name of model's architecture for feature extraction
        """
        return self._set(featureExtractorType=value)

    def setImageStd(self, value):
        """

        Parameters
        ----------
        value : List[float]
            The sequence of standard deviations for each channel, to be used when normalizing images
        """
        return self._set(imageStd=value)

    def setImageMean(self, value):
        """

        Parameters
        ----------
        value : List[float]
            The sequence of means for each channel, to be used when normalizing images
        """
        return self._set(imageMean=value)

    def setResample(self, value):
        """

        Parameters
        ----------
        value : int
            Resampling filter for resizing. This can be one of `PIL.Image.NEAREST`, `PIL.Image.BILINEAR` or
            `PIL.Image.BICUBIC`. Only has an effect if `do_resize` is set to `True`.
        """
        return self._set(resample=value)

    def setSize(self, value):
        """

        Parameters
        ----------
        value : int
            Resize the input to the given size. If a tuple is provided, it should be (width, height).
        """
        return self._set(size=value)


class HasRescaleFactor:
    doRescale = Param(Params._dummy(), "doRescale",
                      "Whether to rescale the image values by rescaleFactor.",
                      TypeConverters.toBoolean)

    rescaleFactor = Param(Params._dummy(), "rescaleFactor",
                          "Factor to scale the image values",
                          TypeConverters.toFloat)

    def setDoRescale(self, value):
        """Sets Whether to rescale the image values by rescaleFactor, by default `True`.

        Parameters
        ----------
        value : Boolean
            Whether to rescale the image values by rescaleFactor.
        """
        return self._set(doRescale=value)

    def setRescaleFactor(self, value):
        """Sets Factor to scale the image values, by default `1/255.0`.

        Parameters
        ----------
        value : Boolean
            Whether to rescale the image values by rescaleFactor.
        """
        return self._set(rescaleFactor=value)


class HasBatchedAnnotateAudio:
    batchSize = Param(Params._dummy(), "batchSize", "Size of every batch", TypeConverters.toInt)

    def setBatchSize(self, v):
        """Sets batch size.

        Parameters
        ----------
        v : int
            Batch size
        """
        return self._set(batchSize=v)

    def getBatchSize(self):
        """Gets current batch size.

        Returns
        -------
        int
            Current batch size
        """
        return self.getOrDefault(self.batchSize)


class HasAudioFeatureProperties:
    doNormalize = Param(Params._dummy(), "doNormalize",
                        "Whether to normalize the input",
                        TypeConverters.toBoolean)

    returnAttentionMask = Param(Params._dummy(), "returnAttentionMask", "",
                                TypeConverters.toBoolean)

    paddingSide = Param(Params._dummy(), "paddingSide",
                        "",
                        TypeConverters.toString)

    featureSize = Param(Params._dummy(), "featureSize",
                        "",
                        TypeConverters.toInt)

    samplingRate = Param(Params._dummy(), "samplingRate",
                         "",
                         TypeConverters.toInt)

    paddingValue = Param(Params._dummy(), "paddingValue",
                         "",
                         TypeConverters.toFloat)

    def setDoNormalize(self, value):
        """

        Parameters
        ----------
        value : Boolean
            Whether to normalize the input with mean and standard deviation
        """
        return self._set(doNormalize=value)

    def setReturnAttentionMask(self, value):
        """

        Parameters
        ----------
        value : boolean
        """
        return self._set(returnAttentionMask=value)

    def setPaddingSide(self, value):
        """

        Parameters
        ----------
        value : str

        """
        return self._set(paddingSide=value)

    def setFeatureSize(self, value):
        """

        Parameters
        ----------
        value : int

        """
        return self._set(featureSize=value)

    def setSamplingRate(self, value):
        """

        Parameters
        ----------
        value : Int
        """
        return self._set(samplingRate=value)

    def setPaddingValue(self, value):
        """

        Parameters
        ----------
        value : float
        """
        return self._set(paddingValue=value)


class HasEngine:
    engine = Param(Params._dummy(), "engine",
                   "Deep Learning engine used for this model",
                   typeConverter=TypeConverters.toString)

    def getEngine(self):
        """

        Returns
        -------
        str
           Deep Learning engine used for this model"
        """
        return self.getOrDefault(self.engine)


class HasCandidateLabelsProperties:
    candidateLabels = Param(Params._dummy(), "candidateLabels",
                            "Deep Learning engine used for this model",
                            typeConverter=TypeConverters.toListString)

    contradictionIdParam = Param(Params._dummy(), "contradictionIdParam",
                                 "contradictionIdParam",
                                 typeConverter=TypeConverters.toInt)

    entailmentIdParam = Param(Params._dummy(), "entailmentIdParam",
                              "contradictionIdParam",
                              typeConverter=TypeConverters.toInt)

    def setCandidateLabels(self, v):
        """Sets candidateLabels.

        Parameters
        ----------
        v : list[string]
            candidateLabels
        """
        return self._set(candidateLabels=v)

    def setContradictionIdParam(self, v):
        """Sets contradictionIdParam.

        Parameters
        ----------
        v : int
            contradictionIdParam
        """
        return self._set(contradictionIdParam=v)

    def setEntailmentIdParam(self, v):
        """Sets entailmentIdParam.

        Parameters
        ----------
        v : int
            entailmentIdParam
        """
        return self._set(entailmentIdParam=v)


class HasMaxSentenceLengthLimit:
    # Default Value, can be overridden
    max_length_limit = 512

    maxSentenceLength = Param(Params._dummy(),
                              "maxSentenceLength",
                              "Max sentence length to process",
                              typeConverter=TypeConverters.toInt)

    def setMaxSentenceLength(self, value):
        """Sets max sentence length to process.

        Note that a maximum limit exists depending on the model. If you are working with long single
        sequences, consider splitting up the input first with another annotator e.g. SentenceDetector.

        Parameters
        ----------
        value : int
            Max sentence length to process
        """
        if value > self.max_length_limit:
            raise ValueError(
                f"{self.__class__.__name__} models do not support token sequences longer than {self.max_length_limit}.\n"
                f"Consider splitting up the input first with another annotator e.g. SentenceDetector.")
        return self._set(maxSentenceLength=value)

    def getMaxSentenceLength(self):
        """Gets max sentence of the model.

        Returns
        -------
        int
            Max sentence length to process
        """
        return self.getOrDefault("maxSentenceLength")


class HasLongMaxSentenceLengthLimit(HasMaxSentenceLengthLimit):
    max_length_limit = 4096


class HasGeneratorProperties:
    task = Param(Params._dummy(), "task", "Transformer's task, e.g. summarize>", typeConverter=TypeConverters.toString)

    minOutputLength = Param(Params._dummy(), "minOutputLength", "Minimum length of the sequence to be generated",
                            typeConverter=TypeConverters.toInt)

    maxOutputLength = Param(Params._dummy(), "maxOutputLength", "Maximum length of output text",
                            typeConverter=TypeConverters.toInt)

    doSample = Param(Params._dummy(), "doSample", "Whether or not to use sampling; use greedy decoding otherwise",
                     typeConverter=TypeConverters.toBoolean)

    temperature = Param(Params._dummy(), "temperature", "The value used to module the next token probabilities",
                        typeConverter=TypeConverters.toFloat)

    topK = Param(Params._dummy(), "topK",
                 "The number of highest probability vocabulary tokens to keep for top-k-filtering",
                 typeConverter=TypeConverters.toInt)

    topP = Param(Params._dummy(), "topP",
                 "If set to float < 1, only the most probable tokens with probabilities that add up to ``top_p`` or higher are kept for generation",
                 typeConverter=TypeConverters.toFloat)

    repetitionPenalty = Param(Params._dummy(), "repetitionPenalty",
                              "The parameter for repetition penalty. 1.0 means no penalty. See `this paper <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details",
                              typeConverter=TypeConverters.toFloat)

    noRepeatNgramSize = Param(Params._dummy(), "noRepeatNgramSize",
                              "If set to int > 0, all ngrams of that size can only occur once",
                              typeConverter=TypeConverters.toInt)

    beamSize = Param(Params._dummy(), "beamSize",
                     "The Number of beams for beam search.",
                     typeConverter=TypeConverters.toInt)

    nReturnSequences = Param(Params._dummy(),
                             "nReturnSequences",
                             "The number of sequences to return from the beam search.",
                             typeConverter=TypeConverters.toInt)

    def setTask(self, value):
        """Sets the transformer's task, e.g. ``summarize:``.

        Parameters
        ----------
        value : str
            The transformer's task
        """
        return self._set(task=value)

    def setMinOutputLength(self, value):
        """Sets minimum length of the sequence to be generated.

        Parameters
        ----------
        value : int
            Minimum length of the sequence to be generated
        """
        return self._set(minOutputLength=value)

    def setMaxOutputLength(self, value):
        """Sets maximum length of output text.

        Parameters
        ----------
        value : int
            Maximum length of output text
        """
        return self._set(maxOutputLength=value)

    def setDoSample(self, value):
        """Sets whether or not to use sampling, use greedy decoding otherwise.

        Parameters
        ----------
        value : bool
            Whether or not to use sampling; use greedy decoding otherwise
        """
        return self._set(doSample=value)

    def setTemperature(self, value):
        """Sets the value used to module the next token probabilities.

        Parameters
        ----------
        value : float
            The value used to module the next token probabilities
        """
        return self._set(temperature=value)

    def setTopK(self, value):
        """Sets the number of highest probability vocabulary tokens to keep for
        top-k-filtering.

        Parameters
        ----------
        value : int
            Number of highest probability vocabulary tokens to keep
        """
        return self._set(topK=value)

    def setTopP(self, value):
        """Sets the top cumulative probability for vocabulary tokens.

        If set to float < 1, only the most probable tokens with probabilities
        that add up to ``topP`` or higher are kept for generation.

        Parameters
        ----------
        value : float
            Cumulative probability for vocabulary tokens
        """
        return self._set(topP=value)

    def setRepetitionPenalty(self, value):
        """Sets the parameter for repetition penalty. 1.0 means no penalty.

        Parameters
        ----------
        value : float
            The repetition penalty

        References
        ----------
        See `Ctrl: A Conditional Transformer Language Model For Controllable
        Generation <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details.
        """
        return self._set(repetitionPenalty=value)

    def setNoRepeatNgramSize(self, value):
        """Sets size of n-grams that can only occur once.

        If set to int > 0, all ngrams of that size can only occur once.

        Parameters
        ----------
        value : int
            N-gram size can only occur once
        """
        return self._set(noRepeatNgramSize=value)

    def setBeamSize(self, value):
        """Sets the number of beam size for beam search.

        Parameters
        ----------
        value : int
            Number of beam size for beam search
        """
        return self._set(beamSize=value)

    def setNReturnSequences(self, value):
        """Sets the number of sequences to return from the beam search.

        Parameters
        ----------
        value : int
            Number of sequences to return
        """
        return self._set(nReturnSequences=value)


class HasLlamaCppProperties:
    # -------- MODEl PARAMETERS --------
    nThreads = Param(Params._dummy(), "nThreads", "Set the number of threads to use during generation",
                     typeConverter=TypeConverters.toInt)
    # nThreadsDraft = Param(Params._dummy(), "nThreadsDraft", "Set the number of threads to use during draft generation",
    #                       typeConverter=TypeConverters.toInt)
    nThreadsBatch = Param(Params._dummy(), "nThreadsBatch",
                          "Set the number of threads to use during batch and prompt processing",
                          typeConverter=TypeConverters.toInt)
    # nThreadsBatchDraft = Param(Params._dummy(), "nThreadsBatchDraft",
    #                            "Set the number of threads to use during batch and prompt processing",
    #                            typeConverter=TypeConverters.toInt)
    nCtx = Param(Params._dummy(), "nCtx", "Set the size of the prompt context", typeConverter=TypeConverters.toInt)
    nBatch = Param(Params._dummy(), "nBatch",
                   "Set the logical batch size for prompt processing (must be >=32 to use BLAS)",
                   typeConverter=TypeConverters.toInt)
    nUbatch = Param(Params._dummy(), "nUbatch",
                    "Set the physical batch size for prompt processing (must be >=32 to use BLAS)",
                    typeConverter=TypeConverters.toInt)
    nDraft = Param(Params._dummy(), "nDraft", "Set the number of tokens to draft for speculative decoding",
                   typeConverter=TypeConverters.toInt)
    # nChunks = Param(Params._dummy(), "nChunks", "Set the maximal number of chunks to process",
    #                 typeConverter=TypeConverters.toInt)
    # nSequences = Param(Params._dummy(), "nSequences", "Set the number of sequences to decode",
    #                    typeConverter=TypeConverters.toInt)
    # pSplit = Param(Params._dummy(), "pSplit", "Set the speculative decoding split probability",
    #                typeConverter=TypeConverters.toFloat)
    nGpuLayers = Param(Params._dummy(), "nGpuLayers", "Set the number of layers to store in VRAM (-1 - use default)",
                       typeConverter=TypeConverters.toInt)
    nGpuLayersDraft = Param(Params._dummy(), "nGpuLayersDraft",
                            "Set the number of layers to store in VRAM for the draft model (-1 - use default)",
                            typeConverter=TypeConverters.toInt)
    # Set how to split the model across GPUs
    #
    #   - NONE: No GPU split
    #   - LAYER: Split the model across GPUs by layer
    #   - ROW: Split the model across GPUs by rows
    gpuSplitMode = Param(Params._dummy(), "gpuSplitMode", "Set how to split the model across GPUs",
                         typeConverter=TypeConverters.toString)
    mainGpu = Param(Params._dummy(), "mainGpu", "Set the main GPU that is used for scratch and small tensors.",
                    typeConverter=TypeConverters.toInt)
    # tensorSplit = Param(Params._dummy(), "tensorSplit", "Set how split tensors should be distributed across GPUs",
    #                     typeConverter=TypeConverters.toListFloat)
    # grpAttnN = Param(Params._dummy(), "grpAttnN", "Set the group-attention factor", typeConverter=TypeConverters.toInt)
    # grpAttnW = Param(Params._dummy(), "grpAttnW", "Set the group-attention width", typeConverter=TypeConverters.toInt)
    ropeFreqBase = Param(Params._dummy(), "ropeFreqBase", "Set the RoPE base frequency, used by NTK-aware scaling",
                         typeConverter=TypeConverters.toFloat)
    ropeFreqScale = Param(Params._dummy(), "ropeFreqScale",
                          "Set the RoPE frequency scaling factor, expands context by a factor of 1/N",
                          typeConverter=TypeConverters.toFloat)
    yarnExtFactor = Param(Params._dummy(), "yarnExtFactor", "Set the YaRN extrapolation mix factor",
                          typeConverter=TypeConverters.toFloat)
    yarnAttnFactor = Param(Params._dummy(), "yarnAttnFactor", "Set the YaRN scale sqrt(t) or attention magnitude",
                           typeConverter=TypeConverters.toFloat)
    yarnBetaFast = Param(Params._dummy(), "yarnBetaFast", "Set the YaRN low correction dim or beta",
                         typeConverter=TypeConverters.toFloat)
    yarnBetaSlow = Param(Params._dummy(), "yarnBetaSlow", "Set the YaRN high correction dim or alpha",
                         typeConverter=TypeConverters.toFloat)
    yarnOrigCtx = Param(Params._dummy(), "yarnOrigCtx", "Set the YaRN original context size of model",
                        typeConverter=TypeConverters.toInt)
    defragmentationThreshold = Param(Params._dummy(), "defragmentationThreshold",
                                     "Set the KV cache defragmentation threshold", typeConverter=TypeConverters.toFloat)
    # Set optimization strategies that help on some NUMA systems (if available)
    #
    # Available Strategies:
    #
    #   - DISABLED: No NUMA optimizations
    #   - DISTRIBUTE: Spread execution evenly over all
    #   - ISOLATE: Only spawn threads on CPUs on the node that execution started on
    #   - NUMA_CTL: Use the CPU map provided by numactl
    #   - MIRROR: Mirrors the model across NUMA nodes
    numaStrategy = Param(Params._dummy(), "numaStrategy",
                         "Set optimization strategies that help on some NUMA systems (if available)",
                         typeConverter=TypeConverters.toString)
    # Set the RoPE frequency scaling method, defaults to linear unless specified by the model.
    #
    #   - NONE: Don't use any scaling
    #   - LINEAR: Linear scaling
    #   - YARN: YaRN RoPE scaling
    ropeScalingType = Param(Params._dummy(), "ropeScalingType",
                            "Set the RoPE frequency scaling method, defaults to linear unless specified by the model",
                            typeConverter=TypeConverters.toString)
    # Set the pooling type for embeddings, use model default if unspecified
    #
    #   - MEAN: Mean Pooling
    #   - CLS: CLS Pooling
    #   - LAST: Last token pooling
    #   - RANK: For reranked models
    poolingType = Param(Params._dummy(), "poolingType",
                        "Set the pooling type for embeddings, use model default if unspecified",
                        typeConverter=TypeConverters.toString)
    modelDraft = Param(Params._dummy(), "modelDraft", "Set the draft model for speculative decoding",
                       typeConverter=TypeConverters.toString)
    modelAlias = Param(Params._dummy(), "modelAlias", "Set a model alias", typeConverter=TypeConverters.toString)
    # lookupCacheStaticFilePath = Param(Params._dummy(), "lookupCacheStaticFilePath",
    #                                   "Set path to static lookup cache to use for lookup decoding (not updated by generation)",
    #                                   typeConverter=TypeConverters.toString)
    # lookupCacheDynamicFilePath = Param(Params._dummy(), "lookupCacheDynamicFilePath",
    #                                    "Set path to dynamic lookup cache to use for lookup decoding (updated by generation)",
    #                                    typeConverter=TypeConverters.toString)
    # loraAdapters = new StructFeature[Map[String, Float]](this, "loraAdapters")
    embedding = Param(Params._dummy(), "embedding", "Whether to load model with embedding support",
                      typeConverter=TypeConverters.toBoolean)
    flashAttention = Param(Params._dummy(), "flashAttention", "Whether to enable Flash Attention",
                           typeConverter=TypeConverters.toBoolean)
    # inputPrefixBos = Param(Params._dummy(), "inputPrefixBos",
    #                        "Whether to add prefix BOS to user inputs, preceding the `--in-prefix` string",
    #                        typeConverter=TypeConverters.toBoolean)
    useMmap = Param(Params._dummy(), "useMmap",
                    "Whether to use memory-map model (faster load but may increase pageouts if not using mlock)",
                    typeConverter=TypeConverters.toBoolean)
    useMlock = Param(Params._dummy(), "useMlock",
                     "Whether to force the system to keep model in RAM rather than swapping or compressing",
                     typeConverter=TypeConverters.toBoolean)
    noKvOffload = Param(Params._dummy(), "noKvOffload", "Whether to disable KV offload",
                        typeConverter=TypeConverters.toBoolean)
    systemPrompt = Param(Params._dummy(), "systemPrompt", "Set a system prompt to use",
                         typeConverter=TypeConverters.toString)
    chatTemplate = Param(Params._dummy(), "chatTemplate", "The chat template to use",
                         typeConverter=TypeConverters.toString)
    logVerbosity = Param(Params._dummy(), "logVerbosity", "Set the log verbosity level",
                         typeConverter=TypeConverters.toInt)
    disableLog = Param(Params._dummy(), "disableLog", "Whether to disable logging",
                       typeConverter=TypeConverters.toBoolean)
    reasoningBudget = Param(Params._dummy(), "reasoningBudget",
                            ("Controls the amount of thinking allowed; currently only one of: -1 for unrestricted "
                            "thinking budget, or 0 to disable thinking (default: -1)"),
                            typeConverter=TypeConverters.toInt)

    # -------- INFERENCE PARAMETERS --------
    inputPrefix = Param(Params._dummy(), "inputPrefix", "Set the prompt to start generation with",
                        typeConverter=TypeConverters.toString)
    inputSuffix = Param(Params._dummy(), "inputSuffix", "Set a suffix for infilling",
                        typeConverter=TypeConverters.toString)
    cachePrompt = Param(Params._dummy(), "cachePrompt", "Whether to remember the prompt to avoid reprocessing it",
                        typeConverter=TypeConverters.toBoolean)
    nPredict = Param(Params._dummy(), "nPredict", "Set the number of tokens to predict",
                     typeConverter=TypeConverters.toInt)
    topK = Param(Params._dummy(), "topK", "Set top-k sampling", typeConverter=TypeConverters.toInt)
    topP = Param(Params._dummy(), "topP", "Set top-p sampling", typeConverter=TypeConverters.toFloat)
    minP = Param(Params._dummy(), "minP", "Set min-p sampling", typeConverter=TypeConverters.toFloat)
    tfsZ = Param(Params._dummy(), "tfsZ", "Set tail free sampling, parameter z", typeConverter=TypeConverters.toFloat)
    typicalP = Param(Params._dummy(), "typicalP", "Set locally typical sampling, parameter p",
                     typeConverter=TypeConverters.toFloat)
    temperature = Param(Params._dummy(), "temperature", "Set the temperature", typeConverter=TypeConverters.toFloat)
    dynamicTemperatureRange = Param(Params._dummy(), "dynatempRange", "Set the dynamic temperature range",
                                    typeConverter=TypeConverters.toFloat)
    dynamicTemperatureExponent = Param(Params._dummy(), "dynatempExponent", "Set the dynamic temperature exponent",
                                       typeConverter=TypeConverters.toFloat)
    repeatLastN = Param(Params._dummy(), "repeatLastN", "Set the last n tokens to consider for penalties",
                        typeConverter=TypeConverters.toInt)
    repeatPenalty = Param(Params._dummy(), "repeatPenalty", "Set the penalty of repeated sequences of tokens",
                          typeConverter=TypeConverters.toFloat)
    frequencyPenalty = Param(Params._dummy(), "frequencyPenalty", "Set the repetition alpha frequency penalty",
                             typeConverter=TypeConverters.toFloat)
    presencePenalty = Param(Params._dummy(), "presencePenalty", "Set the repetition alpha presence penalty",
                            typeConverter=TypeConverters.toFloat)
    miroStat = Param(Params._dummy(), "miroStat", "Set MiroStat sampling strategies.",
                     typeConverter=TypeConverters.toString)
    miroStatTau = Param(Params._dummy(), "mirostatTau", "Set the MiroStat target entropy, parameter tau",
                        typeConverter=TypeConverters.toFloat)
    miroStatEta = Param(Params._dummy(), "mirostatEta", "Set the MiroStat learning rate, parameter eta",
                        typeConverter=TypeConverters.toFloat)
    penalizeNl = Param(Params._dummy(), "penalizeNl", "Whether to penalize newline tokens",
                       typeConverter=TypeConverters.toBoolean)
    nKeep = Param(Params._dummy(), "nKeep", "Set the number of tokens to keep from the initial prompt",
                  typeConverter=TypeConverters.toInt)
    seed = Param(Params._dummy(), "seed", "Set the RNG seed", typeConverter=TypeConverters.toInt)
    nProbs = Param(Params._dummy(), "nProbs", "Set the amount top tokens probabilities to output if greater than 0.",
                   typeConverter=TypeConverters.toInt)
    minKeep = Param(Params._dummy(), "minKeep",
                    "Set the amount of tokens the samplers should return at least (0 = disabled)",
                    typeConverter=TypeConverters.toInt)
    grammar = Param(Params._dummy(), "grammar", "Set BNF-like grammar to constrain generations",
                    typeConverter=TypeConverters.toString)
    penaltyPrompt = Param(Params._dummy(), "penaltyPrompt",
                          "Override which part of the prompt is penalized for repetition.",
                          typeConverter=TypeConverters.toString)
    ignoreEos = Param(Params._dummy(), "ignoreEos",
                      "Set whether to ignore end of stream token and continue generating (implies --logit-bias 2-inf)",
                      typeConverter=TypeConverters.toBoolean)
    disableTokenIds = Param(Params._dummy(), "disableTokenIds", "Set the token ids to disable in the completion",
                            typeConverter=TypeConverters.toListInt)
    stopStrings = Param(Params._dummy(), "stopStrings", "Set strings upon seeing which token generation is stopped",
                        typeConverter=TypeConverters.toListString)
    samplers = Param(Params._dummy(), "samplers", "Set which samplers to use for token generation in the given order",
                     typeConverter=TypeConverters.toListString)
    useChatTemplate = Param(Params._dummy(), "useChatTemplate",
                            "Set whether or not generate should apply a chat template",
                            typeConverter=TypeConverters.toBoolean)

    # -------- MODEL SETTERS --------
    def setNThreads(self, nThreads: int):
        """Set the number of threads to use during generation"""
        return self._set(nThreads=nThreads)

    # def setNThreadsDraft(self, nThreadsDraft: int):
    #     """Set the number of threads to use during draft generation"""
    #     return self._set(nThreadsDraft=nThreadsDraft)

    def setNThreadsBatch(self, nThreadsBatch: int):
        """Set the number of threads to use during batch and prompt processing"""
        return self._set(nThreadsBatch=nThreadsBatch)

    # def setNThreadsBatchDraft(self, nThreadsBatchDraft: int):
    #     """Set the number of threads to use during batch and prompt processing"""
    #     return self._set(nThreadsBatchDraft=nThreadsBatchDraft)

    def setNCtx(self, nCtx: int):
        """Set the size of the prompt context"""
        return self._set(nCtx=nCtx)

    def setNBatch(self, nBatch: int):
        """Set the logical batch size for prompt processing (must be >=32 to use BLAS)"""
        return self._set(nBatch=nBatch)

    def setNUbatch(self, nUbatch: int):
        """Set the physical batch size for prompt processing (must be >=32 to use BLAS)"""
        return self._set(nUbatch=nUbatch)

    def setNDraft(self, nDraft: int):
        """Set the number of tokens to draft for speculative decoding"""
        return self._set(nDraft=nDraft)

    # def setNChunks(self, nChunks: int):
    #     """Set the maximal number of chunks to process"""
    #     return self._set(nChunks=nChunks)

    # def setNSequences(self, nSequences: int):
    #     """Set the number of sequences to decode"""
    #     return self._set(nSequences=nSequences)

    # def setPSplit(self, pSplit: float):
    #     """Set the speculative decoding split probability"""
    #     return self._set(pSplit=pSplit)

    def setNGpuLayers(self, nGpuLayers: int):
        """Set the number of layers to store in VRAM (-1 - use default)"""
        return self._set(nGpuLayers=nGpuLayers)

    def setNGpuLayersDraft(self, nGpuLayersDraft: int):
        """Set the number of layers to store in VRAM for the draft model (-1 - use default)"""
        return self._set(nGpuLayersDraft=nGpuLayersDraft)

    def setGpuSplitMode(self, gpuSplitMode: str):
        """Set how to split the model across GPUs"""
        return self._set(gpuSplitMode=gpuSplitMode)

    def setMainGpu(self, mainGpu: int):
        """Set the main GPU that is used for scratch and small tensors."""
        return self._set(mainGpu=mainGpu)

    # def setTensorSplit(self, tensorSplit: List[float]):
    #     """Set how split tensors should be distributed across GPUs"""
    #     return self._set(tensorSplit=tensorSplit)

    # def setGrpAttnN(self, grpAttnN: int):
    #     """Set the group-attention factor"""
    #     return self._set(grpAttnN=grpAttnN)

    # def setGrpAttnW(self, grpAttnW: int):
    #     """Set the group-attention width"""
    #     return self._set(grpAttnW=grpAttnW)

    def setRopeFreqBase(self, ropeFreqBase: float):
        """Set the RoPE base frequency, used by NTK-aware scaling"""
        return self._set(ropeFreqBase=ropeFreqBase)

    def setRopeFreqScale(self, ropeFreqScale: float):
        """Set the RoPE frequency scaling factor, expands context by a factor of 1/N"""
        return self._set(ropeFreqScale=ropeFreqScale)

    def setYarnExtFactor(self, yarnExtFactor: float):
        """Set the YaRN extrapolation mix factor"""
        return self._set(yarnExtFactor=yarnExtFactor)

    def setYarnAttnFactor(self, yarnAttnFactor: float):
        """Set the YaRN scale sqrt(t) or attention magnitude"""
        return self._set(yarnAttnFactor=yarnAttnFactor)

    def setYarnBetaFast(self, yarnBetaFast: float):
        """Set the YaRN low correction dim or beta"""
        return self._set(yarnBetaFast=yarnBetaFast)

    def setYarnBetaSlow(self, yarnBetaSlow: float):
        """Set the YaRN high correction dim or alpha"""
        return self._set(yarnBetaSlow=yarnBetaSlow)

    def setYarnOrigCtx(self, yarnOrigCtx: int):
        """Set the YaRN original context size of model"""
        return self._set(yarnOrigCtx=yarnOrigCtx)

    def setDefragmentationThreshold(self, defragmentationThreshold: float):
        """Set the KV cache defragmentation threshold"""
        return self._set(defragmentationThreshold=defragmentationThreshold)

    def setNumaStrategy(self, numaStrategy: str):
        """Set optimization strategies that help on some NUMA systems (if available)

        Possible values:

        - DISABLED: No NUMA optimizations
        - DISTRIBUTE: spread execution evenly over all
        - ISOLATE: only spawn threads on CPUs on the node that execution started on
        - NUMA_CTL: use the CPU map provided by numactl
        - MIRROR: Mirrors the model across NUMA nodes
        """
        numaUpper = numaStrategy.upper()
        numaStrategies = ["DISABLED", "DISTRIBUTE", "ISOLATE", "NUMA_CTL", "MIRROR"]
        if numaUpper not in numaStrategies:
            raise ValueError(
                f"Invalid NUMA strategy: {numaUpper}. "
                + f"Valid values are: {numaStrategies}"
            )
        return self._set(numaStrategy=numaStrategy)

    def setRopeScalingType(self, ropeScalingType: str):
        """Set the RoPE frequency scaling method, defaults to linear unless specified by the model.

        Possible values:

        - NONE: Don't use any scaling
        - LINEAR: Linear scaling
        - YARN: YaRN RoPE scaling
        """
        ropeScalingTypeUpper = ropeScalingType.upper()
        ropeScalingTypes = ["NONE", "LINEAR", "YARN"]
        if ropeScalingTypeUpper not in ropeScalingTypes:
            raise ValueError(
                f"Invalid RoPE scaling type: {ropeScalingType}. "
                + f"Valid values are: {ropeScalingTypes}"
            )
        return self._set(ropeScalingType=ropeScalingTypeUpper)

    def setPoolingType(self, poolingType: str):
        """Set the pooling type for embeddings, use model default if unspecified

        Possible values:

        - MEAN: Mean Pooling
        - CLS: CLS Pooling
        - LAST: Last token pooling
        - RANK: For reranked models
        """
        poolingTypeUpper = poolingType.upper()
        poolingTypes = ["NONE", "MEAN", "CLS", "LAST", "RANK"]
        if poolingTypeUpper not in poolingTypes:
            raise ValueError(
                f"Invalid pooling type: {poolingType}. "
                + f"Valid values are: {poolingTypes}"
            )
        return self._set(poolingType=poolingType)

    def setModelDraft(self, modelDraft: str):
        """Set the draft model for speculative decoding"""
        return self._set(modelDraft=modelDraft)

    def setModelAlias(self, modelAlias: str):
        """Set a model alias"""
        return self._set(modelAlias=modelAlias)

    # def setLookupCacheStaticFilePath(self, lookupCacheStaticFilePath: str):
    #     """Set path to static lookup cache to use for lookup decoding (not updated by generation)"""
    #     return self._set(lookupCacheStaticFilePath=lookupCacheStaticFilePath)

    # def setLookupCacheDynamicFilePath(self, lookupCacheDynamicFilePath: str):
    #     """Set path to dynamic lookup cache to use for lookup decoding (updated by generation)"""
    #     return self._set(lookupCacheDynamicFilePath=lookupCacheDynamicFilePath)

    def setFlashAttention(self, flashAttention: bool):
        """Whether to enable Flash Attention"""
        return self._set(flashAttention=flashAttention)

    # def setInputPrefixBos(self, inputPrefixBos: bool):
    #     """Whether to add prefix BOS to user inputs, preceding the `--in-prefix` string"""
    #     return self._set(inputPrefixBos=inputPrefixBos)

    def setUseMmap(self, useMmap: bool):
        """Whether to use memory-map model (faster load but may increase pageouts if not using mlock)"""
        return self._set(useMmap=useMmap)

    def setUseMlock(self, useMlock: bool):
        """Whether to force the system to keep model in RAM rather than swapping or compressing"""
        return self._set(useMlock=useMlock)

    def setNoKvOffload(self, noKvOffload: bool):
        """Whether to disable KV offload"""
        return self._set(noKvOffload=noKvOffload)

    def setSystemPrompt(self, systemPrompt: str):
        """Set a system prompt to use"""
        return self._set(systemPrompt=systemPrompt)

    def setChatTemplate(self, chatTemplate: str):
        """The chat template to use"""
        return self._set(chatTemplate=chatTemplate)

    # -------- INFERENCE SETTERS --------
    def setInputPrefix(self, inputPrefix: str):
        """Set the prompt to start generation with"""
        return self._set(inputPrefix=inputPrefix)

    def setInputSuffix(self, inputSuffix: str):
        """Set a suffix for infilling"""
        return self._set(inputSuffix=inputSuffix)

    def setCachePrompt(self, cachePrompt: bool):
        """Whether to remember the prompt to avoid reprocessing it"""
        return self._set(cachePrompt=cachePrompt)

    def setNPredict(self, nPredict: int):
        """Set the number of tokens to predict"""
        return self._set(nPredict=nPredict)

    def setTopK(self, topK: int):
        """Set top-k sampling"""
        return self._set(topK=topK)

    def setTopP(self, topP: float):
        """Set top-p sampling"""
        return self._set(topP=topP)

    def setMinP(self, minP: float):
        """Set min-p sampling"""
        return self._set(minP=minP)

    def setTfsZ(self, tfsZ: float):
        """Set tail free sampling, parameter z"""
        return self._set(tfsZ=tfsZ)

    def setTypicalP(self, typicalP: float):
        """Set locally typical sampling, parameter p"""
        return self._set(typicalP=typicalP)

    def setTemperature(self, temperature: float):
        """Set the temperature"""
        return self._set(temperature=temperature)

    def setDynamicTemperatureRange(self, dynamicTemperatureRange: float):
        """Set the dynamic temperature range"""
        return self._set(dynamicTemperatureRange=dynamicTemperatureRange)

    def setDynamicTemperatureExponent(self, dynamicTemperatureExponent: float):
        """Set the dynamic temperature exponent"""
        return self._set(dynamicTemperatureExponent=dynamicTemperatureExponent)

    def setRepeatLastN(self, repeatLastN: int):
        """Set the last n tokens to consider for penalties"""
        return self._set(repeatLastN=repeatLastN)

    def setRepeatPenalty(self, repeatPenalty: float):
        """Set the penalty of repeated sequences of tokens"""
        return self._set(repeatPenalty=repeatPenalty)

    def setFrequencyPenalty(self, frequencyPenalty: float):
        """Set the repetition alpha frequency penalty"""
        return self._set(frequencyPenalty=frequencyPenalty)

    def setPresencePenalty(self, presencePenalty: float):
        """Set the repetition alpha presence penalty"""
        return self._set(presencePenalty=presencePenalty)

    def setMiroStat(self, miroStat: str):
        """Set MiroStat sampling strategies."""
        return self._set(miroStat=miroStat)

    def setMiroStatTau(self, miroStatTau: float):
        """Set the MiroStat target entropy, parameter tau"""
        return self._set(miroStatTau=miroStatTau)

    def setMiroStatEta(self, miroStatEta: float):
        """Set the MiroStat learning rate, parameter eta"""
        return self._set(miroStatEta=miroStatEta)

    def setPenalizeNl(self, penalizeNl: bool):
        """Whether to penalize newline tokens"""
        return self._set(penalizeNl=penalizeNl)

    def setNKeep(self, nKeep: int):
        """Set the number of tokens to keep from the initial prompt"""
        return self._set(nKeep=nKeep)

    def setSeed(self, seed: int):
        """Set the RNG seed"""
        return self._set(seed=seed)

    def setNProbs(self, nProbs: int):
        """Set the amount top tokens probabilities to output if greater than 0."""
        return self._set(nProbs=nProbs)

    def setMinKeep(self, minKeep: int):
        """Set the amount of tokens the samplers should return at least (0 = disabled)"""
        return self._set(minKeep=minKeep)

    def setGrammar(self, grammar: str):
        """Set BNF-like grammar to constrain generations"""
        return self._set(grammar=grammar)

    def setPenaltyPrompt(self, penaltyPrompt: str):
        """Override which part of the prompt is penalized for repetition."""
        return self._set(penaltyPrompt=penaltyPrompt)

    def setIgnoreEos(self, ignoreEos: bool):
        """Set whether to ignore end of stream token and continue generating (implies --logit-bias 2-inf)"""
        return self._set(ignoreEos=ignoreEos)

    def setDisableTokenIds(self, disableTokenIds: List[int]):
        """Set the token ids to disable in the completion"""
        return self._set(disableTokenIds=disableTokenIds)

    def setStopStrings(self, stopStrings: List[str]):
        """Set strings upon seeing which token generation is stopped"""
        return self._set(stopStrings=stopStrings)

    def setSamplers(self, samplers: List[str]):
        """Set which samplers to use for token generation in the given order"""
        return self._set(samplers=samplers)

    def setUseChatTemplate(self, useChatTemplate: bool):
        """Set whether generate should apply a chat template"""
        return self._set(useChatTemplate=useChatTemplate)

    def setNParallel(self, nParallel: int):
        """Sets the number of parallel processes for decoding. This is an alias for `setBatchSize`."""
        return self.setBatchSize(nParallel)

    def setLogVerbosity(self, logVerbosity: int):
        """Set the log verbosity level"""
        return self._set(logVerbosity=logVerbosity)

    def setReasoningBudget(self, reasoningBudget: int):
        """Controls the amount of thinking allowed; currently only one of: -1 for unrestricted thinking
        budget, or 0 to disable thinking (default: -1)"""
        return self._set(reasoningBudget=reasoningBudget)

    def setDisableLog(self, disableLog: bool):
        """Whether to disable logging"""
        return self._set(disableLog=disableLog)

    # -------- JAVA SETTERS --------
    def setTokenIdBias(self, tokenIdBias: Dict[int, float]):
        """Set token id bias"""
        return self._call_java("setTokenIdBias", tokenIdBias)

    def setTokenBias(self, tokenBias: Dict[str, float]):
        """Set token id bias"""
        return self._call_java("setTokenBias", tokenBias)

    # def setLoraAdapters(self, loraAdapters: Dict[str, float]):
    #     """Set LoRA adapters with their scaling factors"""
    #     return self._call_java("setLoraAdapters", loraAdapters)

    def getMetadata(self):
        """Gets the metadata of the model"""
        return self._call_java("getMetadata")



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from sparknlp.annotator.cv.vit_for_image_classification import *
from sparknlp.annotator.cv.swin_for_image_classification import *
from sparknlp.annotator.cv.convnext_for_image_classification import *
from sparknlp.annotator.cv.vision_encoder_decoder_for_image_captioning import *
from sparknlp.annotator.cv.clip_for_zero_shot_classification import *
from sparknlp.annotator.cv.blip_for_question_answering import *
from sparknlp.annotator.cv.janus_for_multimodal import *
from sparknlp.annotator.cv.mllama_for_multimodal import *
from sparknlp.annotator.cv.qwen2vl_transformer import *
from sparknlp.annotator.cv.llava_for_multimodal import *
from sparknlp.annotator.cv.phi3_vision_for_multimodal import *
from sparknlp.annotator.cv.smolvlm_transformer import *
from sparknlp.annotator.cv.paligemma_for_multimodal import *
from sparknlp.annotator.cv.gemma3_for_multimodal import *
from sparknlp.annotator.cv.internvl_for_multimodal import *
from sparknlp.annotator.cv.florence2_transformer import *



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the DocumentAssembler."""

from pyspark import keyword_only
from pyspark.ml.param import TypeConverters, Params, Param

from sparknlp.common import AnnotatorType
from sparknlp.internal import AnnotatorTransformer


class DocumentAssembler(AnnotatorTransformer):
    """Prepares data into a format that is processable by Spark NLP.

    This is the entry point for every Spark NLP pipeline. The
    `DocumentAssembler` reads ``String`` columns. Additionally,
    :meth:`.setCleanupMode` can be used to pre-process the
    text (Default: ``disabled``). For possible options please refer the
    parameters section.

    For more extended examples on document pre-processing see the
    `Examples <https://github.com/JohnSnowLabs/spark-nlp/blob/master/examples/python/annotation/text/english/document-assembler/Loading_Documents_With_DocumentAssembler.ipynb>`__.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``NONE``               ``DOCUMENT``
    ====================== ======================

    Parameters
    ----------
    inputCol
        Input column name
    outputCol
        Output column name
    idCol
        Name of String type column for row id.
    metadataCol
        Name of Map type column with metadata information
    cleanupMode
        How to cleanup the document , by default disabled.
        Possible values: ``disabled, inplace, inplace_full, shrink, shrink_full,
        each, each_full, delete_full``

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from pyspark.ml import Pipeline
    >>> data = spark.createDataFrame([["Spark NLP is an open-source text processing library."]]).toDF("text")
    >>> documentAssembler = DocumentAssembler().setInputCol("text").setOutputCol("document")
    >>> result = documentAssembler.transform(data)
    >>> result.select("document").show(truncate=False)
    +----------------------------------------------------------------------------------------------+
    |document                                                                                      |
    +----------------------------------------------------------------------------------------------+
    |[[document, 0, 51, Spark NLP is an open-source text processing library., [sentence -> 0], []]]|
    +----------------------------------------------------------------------------------------------+
    >>> result.select("document").printSchema()
    root
    |-- document: array (nullable = True)
    |    |-- element: struct (containsNull = True)
    |    |    |-- annotatorType: string (nullable = True)
    |    |    |-- begin: integer (nullable = False)
    |    |    |-- end: integer (nullable = False)
    |    |    |-- result: string (nullable = True)
    |    |    |-- metadata: map (nullable = True)
    |    |    |    |-- key: string
    |    |    |    |-- value: string (valueContainsNull = True)
    |    |    |-- embeddings: array (nullable = True)
    |    |    |    |-- element: float (containsNull = False)
    """

    outputAnnotatorType = AnnotatorType.DOCUMENT

    inputCol = Param(Params._dummy(), "inputCol", "input column name", typeConverter=TypeConverters.toString)
    outputCol = Param(Params._dummy(), "outputCol", "output column name", typeConverter=TypeConverters.toString)
    idCol = Param(Params._dummy(), "idCol", "column for setting an id to such string in row", typeConverter=TypeConverters.toString)
    metadataCol = Param(Params._dummy(), "metadataCol", "String to String map column to use as metadata", typeConverter=TypeConverters.toString)
    cleanupMode = Param(Params._dummy(), "cleanupMode", "possible values: disabled, inplace, inplace_full, shrink, shrink_full, each, each_full, delete_full", typeConverter=TypeConverters.toString)
    name = 'DocumentAssembler'

    @keyword_only
    def __init__(self):
        super(DocumentAssembler, self).__init__(classname="com.johnsnowlabs.nlp.DocumentAssembler")
        self._setDefault(outputCol="document", cleanupMode='disabled')

    @keyword_only
    def setParams(self):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setInputCol(self, value):
        """Sets input column name.

        Parameters
        ----------
        value : str
            Name of the input column
        """
        return self._set(inputCol=value)

    def setOutputCol(self, value):
        """Sets output column name.

        Parameters
        ----------
        value : str
            Name of the Output Column
        """
        return self._set(outputCol=value)

    def setIdCol(self, value):
        """Sets name of string type column for row id.

        Parameters
        ----------
        value : str
            Name of the Id Column
        """
        return self._set(idCol=value)

    def setMetadataCol(self, value):
        """Sets name for Map type column with metadata information.

        Parameters
        ----------
        value : str
            Name of the metadata column
        """
        return self._set(metadataCol=value)

    def setCleanupMode(self, value):
        """Sets how to cleanup the document, by default disabled.
        Possible values: ``disabled, inplace, inplace_full, shrink, shrink_full,
        each, each_full, delete_full``

        Parameters
        ----------
        value : str
            Cleanup mode
        """
        if value.strip().lower() not in ['disabled', 'inplace', 'inplace_full', 'shrink', 'shrink_full', 'each', 'each_full', 'delete_full']:
            raise Exception("Cleanup mode possible values: disabled, inplace, inplace_full, shrink, shrink_full, each, each_full, delete_full")
        return self._set(cleanupMode=value)

    def getOutputCol(self):
        """Gets output column name of annotations."""
        return self.getOrDefault(self.outputCol)

    # def getInputCol(self):
    #     """Gets current column names of input annotations."""
    #     return self.getOrDefault(self.inputCol)



#  Copyright 2017-2023 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for OpenAICompletion."""
from sparknlp.common import *

class OpenAICompletion(AnnotatorModel):
    """Transformer that makes a request for OpenAI Completion API for each executor.

   See OpenAI API Doc: https://platform.openai.com/docs/api-reference/completions/create for reference

   ====================== ======================
   Input Annotation types Output Annotation type
   ====================== ======================
   ``DOCUMENT``           ``DOCUMENT``
   ====================== ======================

   Parameters
   ----------
   model
       ID of the OpenAI model to use
   suffix
        The suffix that comes after a completion of inserted text
   maxTokens
        The maximum number of tokens to generate in the completion.
   temperature
        What sampling temperature to use, between 0 and 2
   topP
        An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass
   numberOfCompletions
        How many completions to generate for each prompt.
   logprobs
        Include the log probabilities on the logprobs most likely tokens, as well the chosen tokens.
   echo
       Echo back the prompt in addition to the completion
   stop
      Up to 4 sequences where the API will stop generating further tokens.
   presencePenalty
      Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.
   frequencyPenalty
      Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.
   bestOf
      Generates best_of completions server-side and returns the `best` (the one with the highest log probability per token).
   logitBias
      Modify the likelihood of specified tokens appearing in the completion.
   user
      A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse.

   Examples
   --------
   >>> import sparknlp
   >>> from sparknlp.base import *
   >>> from sparknlp.annotator import *
   >>> from sparknlp.common import *
   >>> from pyspark.ml import Pipeline
   >>> documentAssembler = DocumentAssembler() \\
   ...     .setInputCol("text") \\
   ...     .setOutputCol("document")
   >>> openai_completion = OpenAICompletion() \\
   ...     .setInputCols("document") \\
   ...     .setOutputCol("completion") \\
   ...     .setModel("text-davinci-003") \\
   ...     .setMaxTokens(100)
   >>> pipeline = Pipeline().setStages([
   ...     documentAssembler,
   ...     openai_completion
   ... ])
   >>> empty_df = spark.createDataFrame([[""]], ["text"])
   >>> sample_text= [["Generate a restaurant review."], ["Write a review for a local eatery."], ["Create a JSON with a review of a dining experience."]]
   >>> sample_df = spark.createDataFrame(sample_text).toDF("text")
   >>> sample_df.show()
   +--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
   |completion                                                                                                                                                                                                                                                                                        |
   +--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
   |[{document, 0, 258, \\n\\nI had the pleasure of dining at La Fiorita recently, and it was a truly delightful experience! The menu boasted a wonderful selection of classic Italian dishes, all exquisitely prepared and presented. The service staff was friendly and attentive and really, {}, []}]|
   |[{document, 0, 227, \\n\\nI recently visited Barbecue Joe's for dinner and it was amazing! The menu had so many items to choose from including pulled pork, smoked turkey, brisket, pork ribs, and sandwiches. I opted for the pulled pork sandwich and let, {}, []}]                               |
   |[{document, 0, 172, \\n\\n{ \\n   "review": { \\n      "overallRating": 4, \\n      "reviewBody": "I enjoyed my meal at this restaurant. The food was flavourful, well-prepared and beautifully presented., {}, []}]                                                                                   |
   +--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
   """

    name = "OpenAICompletion"

    inputAnnotatorTypes = [AnnotatorType.DOCUMENT]

    outputAnnotatorType = AnnotatorType.DOCUMENT

    model = Param(Params._dummy(),
                  "model",
                  "ID of the OpenAI model to use",
                  typeConverter=TypeConverters.toString)

    suffix = Param(Params._dummy(),
                   "suffix",
                   "The suffix that comes after a completion of inserted text.",
                   typeConverter=TypeConverters.toString)

    maxTokens = Param(Params._dummy(),
                      "maxTokens",
                      "The maximum number of tokens to generate in the completion.",
                      typeConverter=TypeConverters.toInt)

    temperature = Param(Params._dummy(),
                        "temperature",
                        "What sampling temperature to use, between 0 and 2",
                        typeConverter=TypeConverters.toFloat)

    topP = Param(Params._dummy(),
                      "topP",
                      "An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass",
                      typeConverter=TypeConverters.toFloat)

    numberOfCompletions = Param(Params._dummy(),
                                   "numberOfCompletions",
                                   "How many completions to generate for each prompt.",
                                   typeConverter=TypeConverters.toInt)

    logprobs = Param(Params._dummy(),
                         "logprobs",
                         "Include the log probabilities on the logprobs most likely tokens, as well the chosen tokens.",
                         typeConverter=TypeConverters.toInt)

    echo = Param(Params._dummy(),
                        "echo",
                        "Echo back the prompt in addition to the completion",
                        typeConverter=TypeConverters.toBoolean)

    stop = Param(Params._dummy(),
                  "stop",
                  "Up to 4 sequences where the API will stop generating further tokens.",
                  typeConverter=TypeConverters.toListString)

    presencePenalty = Param(Params._dummy(),
                                 "presencePenalty",
                                 "Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.",
                                 typeConverter=TypeConverters.toFloat)

    frequencyPenalty = Param(Params._dummy(),
                                  "frequencyPenalty",
                                  "Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.",
                                  typeConverter=TypeConverters.toFloat)

    bestOf = Param(Params._dummy(),
                      "bestOf",
                      "Generates best_of completions server-side and returns the `best` (the one with the highest log probability per token).",
                      typeConverter=TypeConverters.toInt)

    logitBias = Param(Params._dummy(),
                  "logitBias",
                  "Modify the likelihood of specified tokens appearing in the completion.",
                  typeConverter=TypeConverters.identity)

    user = Param(Params._dummy(),
                 "user",
                 "A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse.",
                 typeConverter=TypeConverters.toString)

    def setModel(self, value):
        """Sets model ID of the OpenAI model to use

        Parameters
        ----------
        value : str
           ID of the OpenAI model to use
        """
        return self._set(model=value)

    def setSuffix(self, value):
        """Sets the suffix that comes after a completion of inserted text.

        Parameters
        ----------
        value : str
           The suffix that comes after a completion of inserted text.
        """
        return self._set(suffix=value)

    def setMaxTokens(self, value):
        """Sets the maximum number of tokens to generate in the completion.
           The token count of your prompt plus max_tokens cannot exceed the model's context length.

        Parameters
        ----------
        value : int
           The maximum number of tokens to generate in the completion.
        """
        return self._set(maxTokens=value)

    def setTemperature(self, value):
        """Sets what sampling temperature to use, between 0 and 2.
        Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic.
        We generally recommend altering this or top_p but not both.

        Parameters
        ----------
        value : float
           What sampling temperature to use, between 0 and 2
        """
        return self._set(temperature=value)

    def setTopP(self, value):
        """Sets An alternative to sampling with temperature, called nucleus sampling.
        Where the model considers the results of the tokens with top_p probability mass.
        So 0.1 means only the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or temperature but not both.

        Parameters
        ----------
        value : float
           An alternative to sampling with temperature, called nucleus sampling, where the model considers the results
           of the tokens with top_p probability mass
        """
        return self._set(topP=value)

    def setNumberOfCompletions(self, value):
        """Sets how many completions to generate for each prompt.

        Note: Because this parameter generates many completions, it can quickly consume your token quota.
        Use carefully and ensure that you have reasonable settings for max_tokens and stop.

        Parameters
        ----------
        value : int
           How many completions to generate for each prompt.
        """
        return self._set(numberOfCompletions=value)

    def setLogprobs(self, value):
        """Sets the log probabilities on the logprobs most likely tokens, as well the chosen tokens.
        For example, if logprobs is 5, the API will return a list of the 5 most likely tokens.
        The API will always return the logprob of the sampled token, so there may be up to logprobs+1 elements in the response.
        The maximum value for logprobs is 5.

        Parameters
        ----------
        value : int
           How many completions to generate for each prompt.
        """
        return self._set(logprobs=value)

    def setEcho(self, value):
        """Sets echo back the prompt in addition to the completion

        Parameters
        ----------
        value : str
           Echo back the prompt in addition to the completion
        """
        return self._set(echo=value)

    def setStop(self, value):
        """Sets Up to 4 sequences where the API will stop generating further tokens.
        The returned text will not contain the stop sequence.

        Parameters
        ----------
        value : List[str]
           Up to 4 sequences where the API will stop generating further tokens.
        """
        return self._set(stop=value)

    def setPresencePenalty(self, value):
        """Sets values to penalize new tokens based on whether they appear in the text so far,
        increasing the model's likelihood to talk about new topics.

        Parameters
        ----------
        value : int
           Number between -2.0 and 2.0.
           Positive values penalize new tokens based on whether they appear in the text so far,
           increasing the model's likelihood to talk about new topics.
        """
        return self._set(presencePenalty=value)

    def setFrequencyPenalty(self, value):
        """Sets values to penalize new tokens based on their existing frequency in the text so far,
           decreasing the model's likelihood to repeat the same line verbatim.

        Parameters
        ----------
        value : int
           Number between -2.0 and 2.0.
           Positive values penalize new tokens based on their existing frequency in the text so far,
           decreasing the model's likelihood to repeat the same line verbatim.
        """
        return self._set(frequencyPenalty=value)

    def setBestOf(self, value):
        """Sets best_of completions server-side and returns the "best" (the one with the highest log probability per token)

        Parameters
        ----------
        value : int
           Generates best_of completions server-side and returns the "best" (the one with the highest log probability per token).
           Results cannot be streamed.
           When used with n, best_of controls the number of candidate completions and n specifies how many to return – best_of must be greater than n.
           Note: Because this parameter generates many completions, it can quickly consume your token quota.
           Use carefully and ensure that you have reasonable settings for max_tokens and stop.
        """
        return self._set(bestOf=value)

    def setLogitBias(self, value):
        """Sets the likelihood of specified tokens appearing in the completion

        Parameters
        ----------
        value : dict
           Modify the likelihood of specified tokens appearing in the completion.
            Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated
            bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the relevant token.
            As an example, you can pass {"50256": -100} to prevent the <|endoftext|> token from being generated.
        """
        return self._set(logitBias=value)

    def setUser(self, value):
        """Sets a unique identifier representing your end-user

        Parameters
        ----------
        value : str
           A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse.
        """
        return self._set(user=value)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.ml.ai.OpenAICompletion", java_model=None):
        super(OpenAICompletion, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            maxTokens=16,
            temperature=1,
            topP=1,
            numberOfCompletions=1,
            echo=False,
            presencePenalty=0,
            frequencyPenalty=0,
            bestOf=1
        )



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the LLAMA2Transformer."""

from sparknlp.common import *


class LLAMA2Transformer(AnnotatorModel, HasBatchedAnnotate, HasEngine):
    """Llama 2: Open Foundation and Fine-Tuned Chat Models

    The Llama 2 release introduces a family of pretrained and fine-tuned LLMs, ranging in scale
    from 7B to 70B parameters (7B, 13B, 70B). The pretrained models come with significant
    improvements over the Llama 1 models, including being trained on 40% more tokens, having a
    much longer context length (4k tokens 🤯), and using grouped-query attention for fast
    inference of the 70B model🔥!

    However, the most exciting part of this release is the fine-tuned models (Llama 2-Chat), which
    have been optimized for dialogue applications using Reinforcement Learning from Human Feedback
    (RLHF). Across a wide range of helpfulness and safety benchmarks, the Llama 2-Chat models
    perform better than most open models and achieve comparable performance to ChatGPT according
    to human evaluations.

    Pretrained models can be loaded with :meth:`.pretrained` of the companion
    object:

    >>> llama2 = LLAMA2Transformer.pretrained() \\
    ...     .setInputCols(["document"]) \\
    ...     .setOutputCol("generation")


    The default model is ``"llam2-7b"``, if no name is provided. For available
    pretrained models please see the `Models Hub
    <https://sparknlp.org/models?q=llama2>`__.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``DOCUMENT``           ``DOCUMENT``
    ====================== ======================

    Parameters
    ----------
    configProtoBytes
        ConfigProto from tensorflow, serialized into byte array.
    minOutputLength
        Minimum length of the sequence to be generated, by default 0
    maxOutputLength
        Maximum length of output text, by default 20
    doSample
        Whether or not to use sampling; use greedy decoding otherwise, by default False
    temperature
        The value used to module the next token probabilities, by default 1.0
    topK
        The number of highest probability vocabulary tokens to keep for
        top-k-filtering, by default 50
    topP
        Top cumulative probability for vocabulary tokens, by default 1.0

        If set to float < 1, only the most probable tokens with probabilities
        that add up to ``topP`` or higher are kept for generation.
    repetitionPenalty
        The parameter for repetition penalty, 1.0 means no penalty. , by default
        1.0
    noRepeatNgramSize
        If set to int > 0, all ngrams of that size can only occur once, by
        default 0
    ignoreTokenIds
        A list of token ids which are ignored in the decoder's output, by
        default []

    Notes
    -----
    This is a very computationally expensive module especially on larger
    sequence. The use of an accelerator such as GPU is recommended.

    References
    ----------
    - `Llama 2: Open Foundation and Fine-Tuned Chat Models
      <https://ai.meta.com/research/publications/llama-2-open-foundation-and-fine-tuned-chat-models/>`__
    - https://github.com/facebookresearch/llama

    **Paper Abstract:**

    *In this work, we develop and release Llama 2, a collection of pretrained and fine-tuned
    large language models (LLMs) ranging in scale from 7 billion to 70 billion parameters. Our
    fine-tuned LLMs, called Llama 2-Chat, are optimized for dialogue use cases. Our models
    outperform open-source chat models on most benchmarks we tested, and based on our human
    evaluations for helpfulness and safety, may be a suitable substitute for closed-source models.
    We provide a detailed description of our approach to fine-tuning and safety improvements of
    Llama 2-Chat in order to enable the community to build on our work and contribute to the
    responsible development of LLMs.*

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from pyspark.ml import Pipeline
    >>> documentAssembler = DocumentAssembler() \\
    ...     .setInputCol("text") \\
    ...     .setOutputCol("documents")
    >>> llama2 = LLAMA2Transformer.pretrained("llama_2_7b_chat_hf_int4") \\
    ...     .setInputCols(["documents"]) \\
    ...     .setMaxOutputLength(50) \\
    ...     .setOutputCol("generation")
    >>> pipeline = Pipeline().setStages([documentAssembler, llama2])
    >>> data = spark.createDataFrame([["My name is Leonardo."]]).toDF("text")
    >>> result = pipeline.fit(data).transform(data)
    >>> result.select("summaries.generation").show(truncate=False)
    +----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    |result                                                                                                                                                                                              |
    +----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    |[My name is Leonardo. I am a man of letters. I have been a man for many years. I was born in the year 1776. I came to the United States in 1776, and I have lived in the United Kingdom since 1776.]|
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    """

    name = "LLAMA2Transformer"

    inputAnnotatorTypes = [AnnotatorType.DOCUMENT]

    outputAnnotatorType = AnnotatorType.DOCUMENT


    configProtoBytes = Param(Params._dummy(),
                             "configProtoBytes",
                             "ConfigProto from tensorflow, serialized into byte array. Get with config_proto.SerializeToString()",
                             TypeConverters.toListInt)

    minOutputLength = Param(Params._dummy(), "minOutputLength", "Minimum length of the sequence to be generated",
                            typeConverter=TypeConverters.toInt)

    maxOutputLength = Param(Params._dummy(), "maxOutputLength", "Maximum length of output text",
                            typeConverter=TypeConverters.toInt)

    doSample = Param(Params._dummy(), "doSample", "Whether or not to use sampling; use greedy decoding otherwise",
                     typeConverter=TypeConverters.toBoolean)

    temperature = Param(Params._dummy(), "temperature", "The value used to module the next token probabilities",
                        typeConverter=TypeConverters.toFloat)

    topK = Param(Params._dummy(), "topK",
                 "The number of highest probability vocabulary tokens to keep for top-k-filtering",
                 typeConverter=TypeConverters.toInt)

    topP = Param(Params._dummy(), "topP",
                 "If set to float < 1, only the most probable tokens with probabilities that add up to ``top_p`` or higher are kept for generation",
                 typeConverter=TypeConverters.toFloat)

    repetitionPenalty = Param(Params._dummy(), "repetitionPenalty",
                              "The parameter for repetition penalty. 1.0 means no penalty. See `this paper <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details",
                              typeConverter=TypeConverters.toFloat)

    noRepeatNgramSize = Param(Params._dummy(), "noRepeatNgramSize",
                              "If set to int > 0, all ngrams of that size can only occur once",
                              typeConverter=TypeConverters.toInt)

    ignoreTokenIds = Param(Params._dummy(), "ignoreTokenIds",
                           "A list of token ids which are ignored in the decoder's output",
                           typeConverter=TypeConverters.toListInt)


    def setIgnoreTokenIds(self, value):
        """A list of token ids which are ignored in the decoder's output.

        Parameters
        ----------
        value : List[int]
            The words to be filtered out
        """
        return self._set(ignoreTokenIds=value)

    def setConfigProtoBytes(self, b):
        """Sets configProto from tensorflow, serialized into byte array.

        Parameters
        ----------
        b : List[int]
            ConfigProto from tensorflow, serialized into byte array
        """
        return self._set(configProtoBytes=b)

    def setMinOutputLength(self, value):
        """Sets minimum length of the sequence to be generated.

        Parameters
        ----------
        value : int
            Minimum length of the sequence to be generated
        """
        return self._set(minOutputLength=value)

    def setMaxOutputLength(self, value):
        """Sets maximum length of output text.

        Parameters
        ----------
        value : int
            Maximum length of output text
        """
        return self._set(maxOutputLength=value)

    def setDoSample(self, value):
        """Sets whether or not to use sampling, use greedy decoding otherwise.

        Parameters
        ----------
        value : bool
            Whether or not to use sampling; use greedy decoding otherwise
        """
        return self._set(doSample=value)

    def setTemperature(self, value):
        """Sets the value used to module the next token probabilities.

        Parameters
        ----------
        value : float
            The value used to module the next token probabilities
        """
        return self._set(temperature=value)

    def setTopK(self, value):
        """Sets the number of highest probability vocabulary tokens to keep for
        top-k-filtering.

        Parameters
        ----------
        value : int
            Number of highest probability vocabulary tokens to keep
        """
        return self._set(topK=value)

    def setTopP(self, value):
        """Sets the top cumulative probability for vocabulary tokens.

        If set to float < 1, only the most probable tokens with probabilities
        that add up to ``topP`` or higher are kept for generation.

        Parameters
        ----------
        value : float
            Cumulative probability for vocabulary tokens
        """
        return self._set(topP=value)

    def setRepetitionPenalty(self, value):
        """Sets the parameter for repetition penalty. 1.0 means no penalty.

        Parameters
        ----------
        value : float
            The repetition penalty

        References
        ----------
        See `Ctrl: A Conditional Transformer Language Model For Controllable
        Generation <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details.
        """
        return self._set(repetitionPenalty=value)

    def setNoRepeatNgramSize(self, value):
        """Sets size of n-grams that can only occur once.

        If set to int > 0, all ngrams of that size can only occur once.

        Parameters
        ----------
        value : int
            N-gram size can only occur once
        """
        return self._set(noRepeatNgramSize=value)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.annotators.seq2seq.LLAMA2Transformer", java_model=None):
        super(LLAMA2Transformer, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            minOutputLength=0,
            maxOutputLength=20,
            doSample=False,
            temperature=0.6,
            topK=50,
            topP=0.9,
            repetitionPenalty=1.0,
            noRepeatNgramSize=0,
            ignoreTokenIds=[],
            batchSize=1
        )

    @staticmethod
    def loadSavedModel(folder, spark_session, use_openvino = False):
        """Loads a locally saved model.

        Parameters
        ----------
        folder : str
            Folder of the saved model
        spark_session : pyspark.sql.SparkSession
            The current SparkSession

        Returns
        -------
        LLAMA2Transformer
            The restored model
        """
        from sparknlp.internal import _LLAMA2Loader
        jModel = _LLAMA2Loader(folder, spark_session._jsparkSession, use_openvino)._java_obj
        return LLAMA2Transformer(java_model=jModel)

    @staticmethod
    def pretrained(name="llama_2_7b_chat_hf_int4", lang="en", remote_loc=None):
        """Downloads and loads a pretrained model.

        Parameters
        ----------
        name : str, optional
            Name of the pretrained model, by default "llama_2_7b_chat_hf_int4"
        lang : str, optional
            Language of the pretrained model, by default "en"
        remote_loc : str, optional
            Optional remote address of the resource, by default None. Will use
            Spark NLPs repositories otherwise.

        Returns
        -------
        LLAMA2Transformer
            The restored model
        """
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(LLAMA2Transformer, name, lang, remote_loc)



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Python wrapper for the Block GRU Op."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras.engine import input_spec
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import gen_rnn_ops
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import nn_ops
from tensorflow.python.ops import rnn_cell_impl
from tensorflow.python.util.deprecation import deprecated_args

LayerRNNCell = rnn_cell_impl.LayerRNNCell  # pylint: disable=invalid-name


@ops.RegisterGradient("GRUBlockCell")
def _GRUBlockCellGrad(op, *grad):
    r"""Gradient for GRUBlockCell.

    Args:
      op: Op for which the gradient is defined.
      *grad: Gradients of the optimization function wrt output
        for the Op.

    Returns:
      d_x: Gradients wrt to x
      d_h: Gradients wrt to h
      d_w_ru: Gradients wrt to w_ru
      d_w_c: Gradients wrt to w_c
      d_b_ru: Gradients wrt to b_ru
      d_b_c: Gradients wrt to b_c

    Mathematics behind the Gradients below:
    ```
    d_c_bar = d_h \circ (1-u) \circ (1-c \circ c)
    d_u_bar = d_h \circ (h-c) \circ u \circ (1-u)

    d_r_bar_u_bar = [d_r_bar d_u_bar]

    [d_x_component_1 d_h_prev_component_1] = d_r_bar_u_bar * w_ru^T

    [d_x_component_2 d_h_prevr] = d_c_bar * w_c^T

    d_x = d_x_component_1 + d_x_component_2

    d_h_prev = d_h_prev_component_1 + d_h_prevr \circ r + u
    ```
    Below calculation is performed in the python wrapper for the Gradients
    (not in the gradient kernel.)
    ```
    d_w_ru = x_h_prevr^T * d_c_bar

    d_w_c = x_h_prev^T * d_r_bar_u_bar

    d_b_ru = sum of d_r_bar_u_bar along axis = 0

    d_b_c = sum of d_c_bar along axis = 0
    ```
    """
    x, h_prev, w_ru, w_c, b_ru, b_c = op.inputs
    r, u, c, _ = op.outputs
    _, _, _, d_h = grad

    d_x, d_h_prev, d_c_bar, d_r_bar_u_bar = gen_rnn_ops.gru_block_cell_grad(
        x, h_prev, w_ru, w_c, b_ru, b_c, r, u, c, d_h)

    x_h_prev = array_ops.concat([x, h_prev], 1)
    d_w_ru = math_ops.matmul(x_h_prev, d_r_bar_u_bar, transpose_a=True)
    d_b_ru = nn_ops.bias_add_grad(d_r_bar_u_bar)

    x_h_prevr = array_ops.concat([x, h_prev * r], 1)
    d_w_c = math_ops.matmul(x_h_prevr, d_c_bar, transpose_a=True)
    d_b_c = nn_ops.bias_add_grad(d_c_bar)

    return d_x, d_h_prev, d_w_ru, d_w_c, d_b_ru, d_b_c


class GRUBlockCell(LayerRNNCell):
    r"""Block GRU cell implementation.

    Deprecated: use GRUBlockCellV2 instead.

    The implementation is based on:  http://arxiv.org/abs/1406.1078
    Computes the GRU cell forward propagation for 1 time step.

    This kernel op implements the following mathematical equations:

    Biases are initialized with:

    * `b_ru` - constant_initializer(1.0)
    * `b_c` - constant_initializer(0.0)

    ```
    x_h_prev = [x, h_prev]

    [r_bar u_bar] = x_h_prev * w_ru + b_ru

    r = sigmoid(r_bar)
    u = sigmoid(u_bar)

    h_prevr = h_prev \circ r

    x_h_prevr = [x h_prevr]

    c_bar = x_h_prevr * w_c + b_c
    c = tanh(c_bar)

    h = (1-u) \circ c + u \circ h_prev
    ```

    """

    @deprecated_args(None, "cell_size is deprecated, use num_units instead",
                     "cell_size")
    def __init__(self,
                 num_units=None,
                 cell_size=None,
                 reuse=None,
                 name="gru_cell"):
        """Initialize the Block GRU cell.

        Args:
          num_units: int, The number of units in the GRU cell.
          cell_size: int, The old (deprecated) name for `num_units`.
          reuse: (optional) boolean describing whether to reuse variables in an
            existing scope.  If not `True`, and the existing scope already has the
            given variables, an error is raised.
          name: String, the name of the layer. Layers with the same name will
            share weights, but to avoid mistakes we require reuse=True in such
            cases.  By default this is "lstm_cell", for variable-name compatibility
            with `tf.compat.v1.nn.rnn_cell.GRUCell`.

        Raises:
          ValueError: if both cell_size and num_units are not None;
            or both are None.
        """
        super(GRUBlockCell, self).__init__(_reuse=reuse, name=name)
        if (cell_size is None) == (num_units is None):
            raise ValueError(
                "Exactly one of num_units or cell_size must be provided.")
        if num_units is None:
            num_units = cell_size
        self._cell_size = num_units
        # Inputs must be 2-dimensional.
        self.input_spec = input_spec.InputSpec(ndim=2)

    @property
    def state_size(self):
        return self._cell_size

    @property
    def output_size(self):
        return self._cell_size

    def build(self, input_shape):
        # Check if the input size exist.
        input_size = tensor_shape.dimension_value(input_shape[1])
        if input_size is None:
            raise ValueError("Expecting input_size to be set.")

        self._gate_kernel = self.add_variable(
            "w_ru", [input_size + self._cell_size, self._cell_size * 2])
        self._gate_bias = self.add_variable(
            "b_ru", [self._cell_size * 2],
            initializer=init_ops.constant_initializer(1.0))
        self._candidate_kernel = self.add_variable(
            "w_c", [input_size + self._cell_size, self._cell_size])
        self._candidate_bias = self.add_variable(
            "b_c", [self._cell_size],
            initializer=init_ops.constant_initializer(0.0))

        self.built = True

    def call(self, inputs, h_prev):
        """GRU cell."""
        # Check cell_size == state_size from h_prev.
        cell_size = h_prev.get_shape().with_rank(2)[1]
        if cell_size != self._cell_size:
            raise ValueError("Shape of h_prev[1] incorrect: cell_size %i vs %s" %
                             (self._cell_size, cell_size))

        _gru_block_cell = gen_rnn_ops.gru_block_cell
        _, _, _, new_h = _gru_block_cell(
            x=inputs,
            h_prev=h_prev,
            w_ru=self._gate_kernel,
            w_c=self._candidate_kernel,
            b_ru=self._gate_bias,
            b_c=self._candidate_bias)

        return new_h, new_h


class GRUBlockCellV2(GRUBlockCell):
    """Temporary GRUBlockCell impl with a different variable naming scheme.

    Only differs from GRUBlockCell by variable names.
    """

    def build(self, input_shape):
        """GRU cell."""
        input_size = tensor_shape.dimension_value(input_shape[1])
        if input_size is None:
            raise ValueError("Expecting input_size to be set.")

        self._gate_kernel = self.add_variable(
            "gates/kernel", [input_size + self._cell_size, self._cell_size * 2])
        self._gate_bias = self.add_variable(
            "gates/bias", [self._cell_size * 2],
            initializer=init_ops.constant_initializer(1.0))
        self._candidate_kernel = self.add_variable(
            "candidate/kernel", [input_size + self._cell_size, self._cell_size])
        self._candidate_bias = self.add_variable(
            "candidate/bias", [self._cell_size],
            initializer=init_ops.constant_initializer(0.0))

        self.built = True



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import unittest

import pytest

from sparknlp.annotator import *
from sparknlp.base import *
from test.util import SparkSessionForTest


@pytest.mark.fast
class DocumentTokenSplitterTestSpec(unittest.TestCase):
    def setUp(self):
        self.data = SparkSessionForTest.spark.createDataFrame(
            [
                [
                    (
                        "All emotions, and that\none particularly, were abhorrent to his cold, precise"
                        " but\nadmirably balanced mind.\n\nHe was, I take it, the most perfect\nreasoning"
                        " and observing machine that the world has seen."
                    )
                ]
            ]
        ).toDF("text")

    def test_run(self):
        df = self.data

        document_assembler = (
            DocumentAssembler().setInputCol("text").setOutputCol("document")
        )

        document_token_splitter = (
            DocumentTokenSplitter()
            .setInputCols("document")
            .setOutputCol("splits")
            .setNumTokens(3)
            .setTokenOverlap(1)
            .setExplodeSplits(True)
            .setTrimWhitespace(True)
        )

        pipeline = Pipeline().setStages([document_assembler, document_token_splitter])

        pipeline_df = pipeline.fit(df).transform(df)

        results = pipeline_df.select("splits").collect()

        splits = [
            row["splits"][0].result.replace("\n\n", " ").replace("\n", " ")
            for row in results
        ]

        expected = [
            "All emotions, and",
            "and that one",
            "one particularly, were",
            "were abhorrent to",
            "to his cold,",
            "cold, precise but",
            "but admirably balanced",
            "balanced mind. He",
            "He was, I",
            "I take it,",
            "it, the most",
            "most perfect reasoning",
            "reasoning and observing",
            "observing machine that",
            "that the world",
            "world has seen.",
        ]

        assert splits == expected



#  Copyright 2017-2023 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the AutoGGUFEmbeddings."""
from sparknlp.common import *


class AutoGGUFEmbeddings(AnnotatorModel, HasBatchedAnnotate):
    """
    Annotator that uses the llama.cpp library to generate text embeddings with large language
    models

    The type of embedding pooling can be set with the `setPoolingType` method. The default is
    `"MEAN"`. The available options are `"NONE"`, `"MEAN"`, `"CLS"`, and `"LAST"`.

    Pretrained models can be loaded with :meth:`.pretrained` of the companion
    object:

    >>> auto_gguf_model = AutoGGUFEmbeddings.pretrained() \\
    ...     .setInputCols(["document"]) \\
    ...     .setOutputCol("embeddings")

    The default model is ``"Qwen3_Embedding_0.6B_Q8_0_gguf"``, if no name is provided.

    For extended examples of usage, see the
    `AutoGGUFEmbeddingsTest <https://github.com/JohnSnowLabs/spark-nlp/tree/master/src/test/scala/com/johnsnowlabs/nlp/embeddings/AutoGGUFEmbeddingsTest.scala>`__
    and the
    `example notebook <https://github.com/JohnSnowLabs/spark-nlp/tree/master/examples/python/llama.cpp/llama.cpp_in_Spark_NLP_AutoGGUFEmbeddings.ipynb>`__.

    For available pretrained models please see the `Models Hub <https://sparknlp.org/models>`__.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``DOCUMENT``           ``SENTENCE_EMBEDDINGS``
    ====================== ======================

    Parameters
    ----------
    nThreads
        Set the number of threads to use during generation
    nThreadsBatch
        Set the number of threads to use during batch and prompt processing
    nCtx
        Set the size of the prompt context
    nBatch
        Set the logical batch size for prompt processing (must be >=32 to use BLAS)
    nUbatch
        Set the physical batch size for prompt processing (must be >=32 to use BLAS)
    nChunks
        Set the maximal number of chunks to process
    nSequences
        Set the number of sequences to decode
    nGpuLayers
        Set the number of layers to store in VRAM (-1 - use default)
    gpuSplitMode
        Set how to split the model across GPUs
    mainGpu
        Set the main GPU that is used for scratch and small tensors.
    tensorSplit
        Set how split tensors should be distributed across GPUs
    grpAttnN
        Set the group-attention factor
    grpAttnW
        Set the group-attention width
    ropeFreqBase
        Set the RoPE base frequency, used by NTK-aware scaling
    ropeFreqScale
        Set the RoPE frequency scaling factor, expands context by a factor of 1/N
    yarnExtFactor
        Set the YaRN extrapolation mix factor
    yarnAttnFactor
        Set the YaRN scale sqrt(t) or attention magnitude
    yarnBetaFast
        Set the YaRN low correction dim or beta
    yarnBetaSlow
        Set the YaRN high correction dim or alpha
    yarnOrigCtx
        Set the YaRN original context size of model
    defragmentationThreshold
        Set the KV cache defragmentation threshold
    numaStrategy
        Set optimization strategies that help on some NUMA systems (if available)
    ropeScalingType
        Set the RoPE frequency scaling method, defaults to linear unless specified by the model
    poolingType
        Set the pooling type for embeddings, use model default if unspecified
    flashAttention
        Whether to enable Flash Attention
    useMmap
        Whether to use memory-map model (faster load but may increase pageouts if not using mlock)
    useMlock
        Whether to force the system to keep model in RAM rather than swapping or compressing
    noKvOffload
        Whether to disable KV offload

    Notes
    -----
    To use GPU inference with this annotator, make sure to use the Spark NLP GPU package and set
    the number of GPU layers with the `setNGpuLayers` method.

    When using larger models, we recommend adjusting GPU usage with `setNCtx` and `setNGpuLayers`
    according to your hardware to avoid out-of-memory errors.

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from pyspark.ml import Pipeline
    >>> document = DocumentAssembler() \\
    ...     .setInputCol("text") \\
    ...     .setOutputCol("document")
    >>> autoGGUFEmbeddings = AutoGGUFEmbeddings.pretrained() \\
    ...     .setInputCols(["document"]) \\
    ...     .setOutputCol("embeddings") \\
    ...     .setBatchSize(4) \\
    ...     .setNGpuLayers(99) \\
    ...     .setPoolingType("MEAN")
    >>> pipeline = Pipeline().setStages([document, autoGGUFEmbeddings])
    >>> data = spark.createDataFrame([["The moons of Jupiter are 77 in total, with 79 confirmed natural satellites and 2 man-made ones."]]).toDF("text")
    >>> result = pipeline.fit(data).transform(data)
    >>> result.select("embeddings.embeddings").show(truncate = False)
    +--------------------------------------------------------------------------------+
    |                                                                      embeddings|
    +--------------------------------------------------------------------------------+
    |[[-0.034486726, 0.07770534, -0.15982522, -0.017873349, 0.013914132, 0.0365736...|
    +--------------------------------------------------------------------------------+
    """

    name = "AutoGGUFEmbeddings"
    inputAnnotatorTypes = [AnnotatorType.DOCUMENT]
    outputAnnotatorType = AnnotatorType.DOCUMENT

    # -------- MODEl PARAMETERS --------
    nThreads = Param(
        Params._dummy(),
        "nThreads",
        "Set the number of threads to use during generation",
        typeConverter=TypeConverters.toInt,
    )
    nThreadsBatch = Param(
        Params._dummy(),
        "nThreadsBatch",
        "Set the number of threads to use during batch and prompt processing",
        typeConverter=TypeConverters.toInt,
    )
    nCtx = Param(
        Params._dummy(),
        "nCtx",
        "Set the size of the prompt context",
        typeConverter=TypeConverters.toInt,
    )
    nBatch = Param(
        Params._dummy(),
        "nBatch",
        "Set the logical batch size for prompt processing (must be >=32 to use BLAS)",
        typeConverter=TypeConverters.toInt,
    )
    nUbatch = Param(
        Params._dummy(),
        "nUbatch",
        "Set the physical batch size for prompt processing (must be >=32 to use BLAS)",
        typeConverter=TypeConverters.toInt,
    )
    nChunks = Param(
        Params._dummy(),
        "nChunks",
        "Set the maximal number of chunks to process",
        typeConverter=TypeConverters.toInt,
    )
    nSequences = Param(
        Params._dummy(),
        "nSequences",
        "Set the number of sequences to decode",
        typeConverter=TypeConverters.toInt,
    )
    nGpuLayers = Param(
        Params._dummy(),
        "nGpuLayers",
        "Set the number of layers to store in VRAM (-1 - use default)",
        typeConverter=TypeConverters.toInt,
    )
    # Set how to split the model across GPUs
    #
    #   - NONE: No GPU split
    #   - LAYER: Split the model across GPUs by layer
    #   - ROW: Split the model across GPUs by rows
    gpuSplitMode = Param(
        Params._dummy(),
        "gpuSplitMode",
        "Set how to split the model across GPUs",
        typeConverter=TypeConverters.toString,
    )
    mainGpu = Param(
        Params._dummy(),
        "mainGpu",
        "Set the main GPU that is used for scratch and small tensors.",
        typeConverter=TypeConverters.toInt,
    )
    tensorSplit = Param(
        Params._dummy(),
        "tensorSplit",
        "Set how split tensors should be distributed across GPUs",
        typeConverter=TypeConverters.toListFloat,
    )
    grpAttnN = Param(
        Params._dummy(),
        "grpAttnN",
        "Set the group-attention factor",
        typeConverter=TypeConverters.toInt,
    )
    grpAttnW = Param(
        Params._dummy(),
        "grpAttnW",
        "Set the group-attention width",
        typeConverter=TypeConverters.toInt,
    )
    ropeFreqBase = Param(
        Params._dummy(),
        "ropeFreqBase",
        "Set the RoPE base frequency, used by NTK-aware scaling",
        typeConverter=TypeConverters.toFloat,
    )
    ropeFreqScale = Param(
        Params._dummy(),
        "ropeFreqScale",
        "Set the RoPE frequency scaling factor, expands context by a factor of 1/N",
        typeConverter=TypeConverters.toFloat,
    )
    yarnExtFactor = Param(
        Params._dummy(),
        "yarnExtFactor",
        "Set the YaRN extrapolation mix factor",
        typeConverter=TypeConverters.toFloat,
    )
    yarnAttnFactor = Param(
        Params._dummy(),
        "yarnAttnFactor",
        "Set the YaRN scale sqrt(t) or attention magnitude",
        typeConverter=TypeConverters.toFloat,
    )
    yarnBetaFast = Param(
        Params._dummy(),
        "yarnBetaFast",
        "Set the YaRN low correction dim or beta",
        typeConverter=TypeConverters.toFloat,
    )
    yarnBetaSlow = Param(
        Params._dummy(),
        "yarnBetaSlow",
        "Set the YaRN high correction dim or alpha",
        typeConverter=TypeConverters.toFloat,
    )
    yarnOrigCtx = Param(
        Params._dummy(),
        "yarnOrigCtx",
        "Set the YaRN original context size of model",
        typeConverter=TypeConverters.toInt,
    )
    defragmentationThreshold = Param(
        Params._dummy(),
        "defragmentationThreshold",
        "Set the KV cache defragmentation threshold",
        typeConverter=TypeConverters.toFloat,
    )
    # Set optimization strategies that help on some NUMA systems (if available)
    #
    # Available Strategies:
    #
    #   - DISABLED: No NUMA optimizations
    #   - DISTRIBUTE: Spread execution evenly over all
    #   - ISOLATE: Only spawn threads on CPUs on the node that execution started on
    #   - NUMA_CTL: Use the CPU map provided by numactl
    #   - MIRROR: Mirrors the model across NUMA nodes
    numaStrategy = Param(
        Params._dummy(),
        "numaStrategy",
        "Set optimization strategies that help on some NUMA systems (if available)",
        typeConverter=TypeConverters.toString,
    )
    # Set the RoPE frequency scaling method, defaults to linear unless specified by the model.
    #
    #   - UNSPECIFIED: Don't use any scaling
    #   - LINEAR: Linear scaling
    #   - YARN: YaRN RoPE scaling
    ropeScalingType = Param(
        Params._dummy(),
        "ropeScalingType",
        "Set the RoPE frequency scaling method, defaults to linear unless specified by the model",
        typeConverter=TypeConverters.toString,
    )
    # Set the pooling type for embeddings, use model default if unspecified
    #
    #   - 0 UNSPECIFIED: Don't use any pooling
    #   - 1 MEAN: Mean Pooling
    #   - 2 CLS: CLS Pooling
    poolingType = Param(
        Params._dummy(),
        "poolingType",
        "Set the pooling type for embeddings, use model default if unspecified",
        typeConverter=TypeConverters.toString,
    )
    flashAttention = Param(
        Params._dummy(),
        "flashAttention",
        "Whether to enable Flash Attention",
        typeConverter=TypeConverters.toBoolean,
    )
    useMmap = Param(
        Params._dummy(),
        "useMmap",
        "Whether to use memory-map model (faster load but may increase pageouts if not using mlock)",
        typeConverter=TypeConverters.toBoolean,
    )
    useMlock = Param(
        Params._dummy(),
        "useMlock",
        "Whether to force the system to keep model in RAM rather than swapping or compressing",
        typeConverter=TypeConverters.toBoolean,
    )
    noKvOffload = Param(
        Params._dummy(),
        "noKvOffload",
        "Whether to disable KV offload",
        typeConverter=TypeConverters.toBoolean,
    )

    # -------- MODEL SETTERS --------
    def setNThreads(self, nThreads: int):
        """Set the number of threads to use during generation"""
        return self._set(nThreads=nThreads)

    def setNThreadsBatch(self, nThreadsBatch: int):
        """Set the number of threads to use during batch and prompt processing"""
        return self._set(nThreadsBatch=nThreadsBatch)

    def setNCtx(self, nCtx: int):
        """Set the size of the prompt context"""
        return self._set(nCtx=nCtx)

    def setNBatch(self, nBatch: int):
        """Set the logical batch size for prompt processing (must be >=32 to use BLAS)"""
        return self._set(nBatch=nBatch)

    def setNUbatch(self, nUbatch: int):
        """Set the physical batch size for prompt processing (must be >=32 to use BLAS)"""
        return self._set(nUbatch=nUbatch)

    def setNChunks(self, nChunks: int):
        """Set the maximal number of chunks to process"""
        return self._set(nChunks=nChunks)

    def setNSequences(self, nSequences: int):
        """Set the number of sequences to decode"""
        return self._set(nSequences=nSequences)

    def setNGpuLayers(self, nGpuLayers: int):
        """Set the number of layers to store in VRAM (-1 - use default)"""
        return self._set(nGpuLayers=nGpuLayers)

    def setGpuSplitMode(self, gpuSplitMode: str):
        """Set how to split the model across GPUs"""
        return self._set(gpuSplitMode=gpuSplitMode)

    def setMainGpu(self, mainGpu: int):
        """Set the main GPU that is used for scratch and small tensors."""
        return self._set(mainGpu=mainGpu)

    def setTensorSplit(self, tensorSplit: List[float]):
        """Set how split tensors should be distributed across GPUs"""
        return self._set(tensorSplit=tensorSplit)

    def setGrpAttnN(self, grpAttnN: int):
        """Set the group-attention factor"""
        return self._set(grpAttnN=grpAttnN)

    def setGrpAttnW(self, grpAttnW: int):
        """Set the group-attention width"""
        return self._set(grpAttnW=grpAttnW)

    def setRopeFreqBase(self, ropeFreqBase: float):
        """Set the RoPE base frequency, used by NTK-aware scaling"""
        return self._set(ropeFreqBase=ropeFreqBase)

    def setRopeFreqScale(self, ropeFreqScale: float):
        """Set the RoPE frequency scaling factor, expands context by a factor of 1/N"""
        return self._set(ropeFreqScale=ropeFreqScale)

    def setYarnExtFactor(self, yarnExtFactor: float):
        """Set the YaRN extrapolation mix factor"""
        return self._set(yarnExtFactor=yarnExtFactor)

    def setYarnAttnFactor(self, yarnAttnFactor: float):
        """Set the YaRN scale sqrt(t) or attention magnitude"""
        return self._set(yarnAttnFactor=yarnAttnFactor)

    def setYarnBetaFast(self, yarnBetaFast: float):
        """Set the YaRN low correction dim or beta"""
        return self._set(yarnBetaFast=yarnBetaFast)

    def setYarnBetaSlow(self, yarnBetaSlow: float):
        """Set the YaRN high correction dim or alpha"""
        return self._set(yarnBetaSlow=yarnBetaSlow)

    def setYarnOrigCtx(self, yarnOrigCtx: int):
        """Set the YaRN original context size of model"""
        return self._set(yarnOrigCtx=yarnOrigCtx)

    def setDefragmentationThreshold(self, defragmentationThreshold: float):
        """Set the KV cache defragmentation threshold"""
        return self._set(defragmentationThreshold=defragmentationThreshold)

    def setNumaStrategy(self, numaStrategy: str):
        """Set optimization strategies that help on some NUMA systems (if available)"""
        numaUpper = numaStrategy.upper()
        numaStrategies = ["DISABLED", "DISTRIBUTE", "ISOLATE", "NUMA_CTL", "MIRROR"]
        if numaUpper not in numaStrategies:
            raise ValueError(
                f"Invalid NUMA strategy: {numaUpper}. "
                + f"Valid values are: {numaStrategies}"
            )
        return self._set(numaStrategy=numaStrategy)

    def setRopeScalingType(self, ropeScalingType: str):
        """Set the RoPE frequency scaling method, defaults to linear unless specified by the model"""
        return self._set(ropeScalingType=ropeScalingType)

    def setPoolingType(self, poolingType: str):
        """Set the pooling type for embeddings, use model default if unspecified"""
        poolingTypeUpper = poolingType.upper()
        poolingTypes = ["NONE", "MEAN", "CLS", "LAST"]
        if poolingTypeUpper not in poolingTypes:
            raise ValueError(
                f"Invalid pooling type: {poolingType}. "
                + f"Valid values are: {poolingTypes}"
            )
        return self._set(poolingType=poolingType)

    def setFlashAttention(self, flashAttention: bool):
        """Whether to enable Flash Attention"""
        return self._set(flashAttention=flashAttention)

    def setUseMmap(self, useMmap: bool):
        """Whether to use memory-map model (faster load but may increase pageouts if not using mlock)"""
        return self._set(useMmap=useMmap)

    def setUseMlock(self, useMlock: bool):
        """Whether to force the system to keep model in RAM rather than swapping or compressing"""
        return self._set(useMlock=useMlock)

    def setNoKvOffload(self, noKvOffload: bool):
        """Whether to disable KV offload"""
        return self._set(noKvOffload=noKvOffload)

    def setNParallel(self, nParallel: int):
        """Sets the number of parallel processes for decoding. This is an alias for `setBatchSize`."""
        return self.setBatchSize(nParallel)

    def getMetadata(self):
        """Gets the metadata of the model"""
        return self._call_java("getMetadata")

    @keyword_only
    def __init__(
            self,
            classname="com.johnsnowlabs.nlp.embeddings.AutoGGUFEmbeddings",
            java_model=None,
    ):
        super(AutoGGUFEmbeddings, self).__init__(
            classname=classname, java_model=java_model
        )
        self._setDefault(
            nCtx=4096,
            nBatch=512,
            poolingType="MEAN",
            nGpuLayers=99,
        )

    @staticmethod
    def loadSavedModel(folder, spark_session):
        """Loads a locally saved model.

        Parameters
        ----------
        folder : str
            Folder of the saved model
        spark_session : pyspark.sql.SparkSession
            The current SparkSession

        Returns
        -------
        AutoGGUFEmbeddings
            The restored model
        """
        from sparknlp.internal import _AutoGGUFEmbeddingsLoader

        jModel = _AutoGGUFEmbeddingsLoader(folder, spark_session._jsparkSession)._java_obj
        return AutoGGUFEmbeddings(java_model=jModel)

    @staticmethod
    def pretrained(name="Qwen3_Embedding_0.6B_Q8_0_gguf", lang="en", remote_loc=None):
        """Downloads and loads a pretrained model.

        Parameters
        ----------
        name : str, optional
            Name of the pretrained model, by default "Qwen3_Embedding_0.6B_Q8_0_gguf"
        lang : str, optional
            Language of the pretrained model, by default "en"
        remote_loc : str, optional
            Optional remote address of the resource, by default None. Will use
            Spark NLPs repositories otherwise.

        Returns
        -------
        AutoGGUFEmbeddings
            The restored model
        """
        from sparknlp.pretrained import ResourceDownloader

        return ResourceDownloader.downloadModel(
            AutoGGUFEmbeddings, name, lang, remote_loc
        )

    def close(self):
        """Closes the llama.cpp model backend freeing resources. The model is reloaded when used again.
        """
        self._java_obj.close()



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Contains the Annotation data format
"""

from pyspark.sql.types import *


class Annotation:
    """Represents the output of Spark NLP Annotators and their details.

    Parameters
    ----------
    annotator_type : str
        The type of the output of the annotator. Possible values are ``DOCUMENT,
        TOKEN, WORDPIECE, WORD_EMBEDDINGS, SENTENCE_EMBEDDINGS, CATEGORY, DATE,
        ENTITY, SENTIMENT, POS, CHUNK, NAMED_ENTITY, NEGEX, DEPENDENCY,
        LABELED_DEPENDENCY, LANGUAGE, KEYWORD, DUMMY``.
    begin : int
        The index of the first character under this annotation.
    end : int
        The index of the last character under this annotation.
    result : str
        The resulting string of the annotation.
    metadata : dict
        Associated metadata for this annotation
    embeddings : list
        Embeddings vector where applicable
    """

    def __init__(self, annotatorType, begin, end, result, metadata, embeddings):
        self.annotatorType = annotatorType
        self.begin = begin
        self.end = end
        self.result = result
        self.metadata = metadata
        self.embeddings = embeddings

    def copy(self, result):
        """Creates new Annotation with a different result, containing all
        settings of this Annotation.

        Parameters
        ----------
        result : str
            The result of the annotation that should be copied.

        Returns
        -------
        Annotation
            Newly created Annotation
        """
        return Annotation(self.annotatorType, self.begin, self.end, result, self.metadata, self.embeddings)

    def __str__(self):
        return "Annotation(%s, %i, %i, %s, %s, %s)" % (
            self.annotatorType,
            self.begin,
            self.end,
            self.result,
            str(self.metadata),
            str(self.embeddings)
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        same_annotator_type = self.annotatorType == other.annotatorType
        same_result = self.result == other.result
        same_begin = self.begin == other.begin
        same_end = self.end == other.end
        same_metadata = dict(self.metadata) == other.metadata
        same_embeddings = self.embeddings == other.embeddings

        same_annotation = \
            same_annotator_type and same_result and same_begin and same_end and same_metadata and same_embeddings

        return same_annotation

    @staticmethod
    def dataType():
        """Returns a Spark `StructType`, that represents the schema of the
        Annotation.

        The Schema looks like::

            struct (containsNull = True)
            |-- annotatorType: string (nullable = False)
            |-- begin: integer (nullable = False)
            |-- end: integer (nullable = False)
            |-- result: string (nullable = False)
            |-- metadata: map (nullable = False)
            |    |-- key: string
            |    |-- value: string (valueContainsNull = True)
            |-- embeddings: array (nullable = False)
            |    |-- element: float (containsNull = False)

        Returns
        -------
        :class:`pyspark.sql.types.StructType`
            Spark Schema of the Annotation
        """
        return StructType([
            StructField('annotatorType', StringType(), False),
            StructField('begin', IntegerType(), False),
            StructField('end', IntegerType(), False),
            StructField('result', StringType(), False),
            StructField('metadata', MapType(StringType(), StringType()), False),
            StructField('embeddings', ArrayType(FloatType()), False)
        ])

    @staticmethod
    def arrayType():
        """Returns a Spark `ArrayType`, that contains the `dataType` of the
        annotation.

        Returns
        -------
        :class:`pyspark.sql.types.ArrayType`
            ArrayType with the Annotation data type embedded.
        """
        return ArrayType(Annotation.dataType())

    @staticmethod
    def fromRow(row):
        """Creates a Annotation from a Spark `Row`.

        Parameters
        ----------
        row : :class:`pyspark.sql.Row`
            Spark row containing columns for ``annotatorType, begin, end,
            result, metadata, embeddings``.

        Returns
        -------
        Annotation
            The new Annotation.
        """
        return Annotation(row.annotatorType, row.begin, row.end, row.result, row.metadata, row.embeddings)

    @staticmethod
    def toRow(annotation):
        """Transforms an Annotation to a Spark `Row`.

        Parameters
        ----------
        annotation : Annotation
            The Annotation to be transformed.

        Returns
        -------
        :class:`pyspark.sql.Row`
            The new Row.
        """
        from pyspark.sql import Row
        return Row(annotation.annotatorType, annotation.begin, annotation.end, annotation.result, annotation.metadata,
                   annotation.embeddings)



#  Copyright 2017-2026 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters

from sparknlp.common import AnnotatorType, AnnotatorProperties
from sparknlp.internal import AnnotatorTransformer


class LayoutAlignerForText(AnnotatorTransformer, AnnotatorProperties):
    """Rebuilds final text by combining aligned document chunks and image captions.

    This transformer is designed to consume ``aligned_doc`` + ``image_caption`` pairs and produce
    coherent text output with re-computed ``begin`` and ``end`` indexes.

    ======================= ======================
    Input Annotation types  Output Annotation type
    ======================= ======================
    ``DOCUMENT, DOCUMENT``  ``DOCUMENT``
    ======================= ======================
    """

    name = "LayoutAlignerForText"

    inputAnnotatorTypes = [AnnotatorType.DOCUMENT, AnnotatorType.DOCUMENT]
    outputAnnotatorType = AnnotatorType.DOCUMENT

    joinDelimiter = Param(
        Params._dummy(),
        "joinDelimiter",
        "Delimiter used to join rebuilt text segments.",
        typeConverter=TypeConverters.toString,
    )

    inlinePrefixThreshold = Param(
        Params._dummy(),
        "inlinePrefixThreshold",
        "Inline images with x <= threshold are inserted before paragraph text.",
        typeConverter=TypeConverters.toInt,
    )

    explodeElements = Param(
        Params._dummy(),
        "explodeElements",
        "Whether to emit one output row per aligned text element.",
        typeConverter=TypeConverters.toBoolean,
    )

    @keyword_only
    def __init__(self):
        super(LayoutAlignerForText, self).__init__(
            classname="com.johnsnowlabs.reader.LayoutAlignerForText"
        )
        self._setDefault(
            outputCol="aligned_text",
            joinDelimiter="\n",
            inlinePrefixThreshold=10,
            explodeElements=False,
        )

    @keyword_only
    def setParams(self):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setJoinDelimiter(self, value):
        return self._set(joinDelimiter=value)

    def setInlinePrefixThreshold(self, value):
        return self._set(inlinePrefixThreshold=value)

    def setExplodeElements(self, value):
        return self._set(explodeElements=value)



#  Copyright 2017-2025 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for Cleaner."""
from sparknlp.annotator import MarianTransformer
from sparknlp.common import *

class Cleaner(MarianTransformer):
    name = "Cleaner"

    inputAnnotatorTypes = [AnnotatorType.TOKEN]

    outputAnnotatorType = AnnotatorType.CHUNK

    encoding = Param(Params._dummy(),
                   "encoding",
                   "The encoding to be used for decoding the byte string (default is utf-8)",
                   typeConverter=TypeConverters.toString)

    cleanPrefixPattern = Param(Params._dummy(),
                     "cleanPrefixPattern",
                     "The pattern for the prefix. Can be a simple string or a regex pattern.",
                     typeConverter=TypeConverters.toString)

    cleanPostfixPattern = Param(Params._dummy(),
                               "cleanPostfixPattern",
                               "The pattern for the postfix. Can be a simple string or a regex pattern.",
                               typeConverter=TypeConverters.toString)

    cleanerMode = Param(
        Params._dummy(),
        "cleanerMode",
        "possible values: " +
        "clean, bytes_string_to_string, clean_non_ascii_chars, clean_ordered_bullets, clean_postfix, clean_prefix, remove_punctuation, replace_unicode_quotes",
        typeConverter=TypeConverters.toString
    )

    extraWhitespace = Param(Params._dummy(),
                    "extraWhitespace",
                    "Whether to remove extra whitespace.",
                    typeConverter=TypeConverters.toBoolean)

    dashes = Param(Params._dummy(),
                "dashes",
                "Whether to handle dashes in text.",
                typeConverter=TypeConverters.toBoolean)

    bullets = Param(Params._dummy(),
                   "bullets",
                   "Whether to handle bullets in text.",
                   typeConverter=TypeConverters.toBoolean)

    trailingPunctuation = Param(Params._dummy(),
                    "trailingPunctuation",
                    "Whether to remove trailing punctuation from text.",
                    typeConverter=TypeConverters.toBoolean)

    lowercase = Param(Params._dummy(),
                "lowercase",
                "Whether to convert text to lowercase.",
                typeConverter=TypeConverters.toBoolean)

    ignoreCase = Param(Params._dummy(),
                      "ignoreCase",
                      "If true, ignores case in the pattern.",
                      typeConverter=TypeConverters.toBoolean)

    strip = Param(Params._dummy(),
               "strip",
               "If true, removes leading or trailing whitespace from the cleaned string.",
               typeConverter=TypeConverters.toBoolean)

    def setEncoding(self, value):
        """Sets the encoding to be used for decoding the byte string (default is utf-8).

        Parameters
        ----------
        value : str
            The encoding to be used for decoding the byte string (default is utf-8)
        """
        return self._set(encoding=value)

    def setCleanPrefixPattern(self, value):
        """Sets the pattern for the prefix. Can be a simple string or a regex pattern.

        Parameters
        ----------
        value : str
            The pattern for the prefix. Can be a simple string or a regex pattern.
        """
        return self._set(cleanPrefixPattern=value)

    def setCleanPostfixPattern(self, value):
        """Sets the pattern for the postfix. Can be a simple string or a regex pattern.

        Parameters
        ----------
        value : str
            The pattern for the postfix. Can be a simple string or a regex pattern.
        """
        return self._set(cleanPostfixPattern=value)

    def setCleanerMode(self, value):
        """Sets the cleaner mode.

        Possible values:
            clean, bytes_string_to_string, clean_non_ascii_chars, clean_ordered_bullets,
            clean_postfix, clean_prefix, remove_punctuation, replace_unicode_quotes

        Parameters
        ----------
        value : str
            The mode for cleaning operations.
        """
        return self._set(cleanerMode=value)

    def setExtraWhitespace(self, value):
        """Sets whether to remove extra whitespace.

        Parameters
        ----------
        value : bool
            Whether to remove extra whitespace.
        """
        return self._set(extraWhitespace=value)

    def setDashes(self, value):
        """Sets whether to handle dashes in text.

        Parameters
        ----------
        value : bool
            Whether to handle dashes in text.
        """
        return self._set(dashes=value)

    def setBullets(self, value):
        """Sets whether to handle bullets in text.

        Parameters
        ----------
        value : bool
            Whether to handle bullets in text.
        """
        return self._set(bullets=value)

    def setTrailingPunctuation(self, value):
        """Sets whether to remove trailing punctuation from text.

        Parameters
        ----------
        value : bool
            Whether to remove trailing punctuation from text.
        """
        return self._set(trailingPunctuation=value)

    def setLowercase(self, value):
        """Sets whether to convert text to lowercase.

        Parameters
        ----------
        value : bool
            Whether to convert text to lowercase.
        """
        return self._set(lowercase=value)

    def setIgnoreCase(self, value):
        """Sets whether to ignore case in the pattern.

        Parameters
        ----------
        value : bool
            If true, ignores case in the pattern.
        """
        return self._set(ignoreCase=value)

    def setStrip(self, value):
        """Sets whether to remove leading or trailing whitespace from the cleaned string.

        Parameters
        ----------
        value : bool
            If true, removes leading or trailing whitespace from the cleaned string.
        """
        return self._set(strip=value)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.annotators.cleaners.Cleaner", java_model=None):
        super(Cleaner, self).__init__(
            classname=classname,
            java_model=java_model
        )


#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains base classes for Recursive Annotators and Estimators."""

from abc import ABC

from pyspark.ml.wrapper import JavaEstimator, JavaModel
from pyspark.sql import DataFrame


class RecursiveEstimator(JavaEstimator, ABC):

    def _fit_java(self, dataset, pipeline=None):
        self._transfer_params_to_java()
        if pipeline:
            return self._java_obj.recursiveFit(dataset._jdf, pipeline._to_java())
        else:
            return self._java_obj.fit(dataset._jdf)

    def _fit(self, dataset, pipeline=None):
        java_model = self._fit_java(dataset, pipeline)
        model = self._create_model(java_model)
        return self._copyValues(model)

    def fit(self, dataset, params=None, pipeline=None):
        if params is None:
            params = dict()
        if isinstance(params, (list, tuple)):
            models = [None] * len(params)
            for index, model in self.fitMultiple(dataset, params):
                models[index] = model
            return models
        elif isinstance(params, dict):
            if params:
                return self.copy(params)._fit(dataset, pipeline=pipeline)
            else:
                return self._fit(dataset, pipeline=pipeline)
        else:
            raise ValueError("Params must be either a param map or a list/tuple of param maps, "
                             "but got %s." % type(params))

class RecursiveTransformer(JavaModel):

    def _transform_recursive(self, dataset, recursive_pipeline):
        self._transfer_params_to_java()
        return DataFrame(self._java_obj.recursiveTransform(dataset._jdf, recursive_pipeline._to_java()),
                         dataset.sql_ctx)

    def transform_recursive(self, dataset, recursive_pipeline, params=None):
        if params is None:
            params = dict()
        if isinstance(params, dict):
            if params:
                return self.copy(params)._transform_recursive(dataset, recursive_pipeline)
            else:
                return self._transform_recursive(dataset, recursive_pipeline)
        else:
            raise ValueError("Params must be a param map but got %s." % type(params))




#  Copyright 2017-2024 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the GGUFRankingFinisher."""

from pyspark import keyword_only
from pyspark.ml.param import TypeConverters, Params, Param
from sparknlp.internal import AnnotatorTransformer


class GGUFRankingFinisher(AnnotatorTransformer):
    """Finisher for AutoGGUFReranker outputs that provides ranking capabilities
    including top-k selection, sorting by relevance score, and score normalization.

    This finisher processes the output of AutoGGUFReranker, which contains documents with
    relevance scores in their metadata. It provides several options for post-processing:

    - Top-k selection: Select only the top k documents by relevance score
    - Score thresholding: Filter documents by minimum relevance score
    - Min-max scaling: Normalize relevance scores to 0-1 range
    - Sorting: Sort documents by relevance score in descending order
    - Ranking: Add rank information to document metadata

    The finisher preserves the document annotation structure while adding ranking information
    to the metadata and optionally filtering/sorting the documents.

    For extended examples of usage, see the `Examples
    <https://github.com/JohnSnowLabs/spark-nlp/blob/master/examples/python/finisher/gguf_ranking_finisher_example.py>`__.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``DOCUMENT``           ``DOCUMENT``
    ====================== ======================

    Parameters
    ----------
    inputCols
        Name of input annotation columns containing reranked documents
    outputCol
        Name of output annotation column containing ranked documents, by default "ranked_documents"
    topK
        Maximum number of top documents to return based on relevance score (-1 for no limit), by default -1
    minRelevanceScore
        Minimum relevance score threshold for filtering documents, by default Double.MinValue
    minMaxScaling
        Whether to apply min-max scaling to normalize relevance scores to 0-1 range, by default False

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from pyspark.ml import Pipeline
    >>> documentAssembler = DocumentAssembler() \\
    ...     .setInputCol("text") \\
    ...     .setOutputCol("document")
    >>> reranker = AutoGGUFReranker.pretrained() \\
    ...     .setInputCols("document") \\
    ...     .setOutputCol("reranked_documents") \\
    ...     .setQuery("A man is eating pasta.")
    >>> finisher = GGUFRankingFinisher() \\
    ...     .setInputCols("reranked_documents") \\
    ...     .setOutputCol("ranked_documents") \\
    ...     .setTopK(3) \\
    ...     .setMinMaxScaling(True)
    >>> pipeline = Pipeline().setStages([documentAssembler, reranker, finisher])
    >>> data = spark.createDataFrame([
    ...     ("A man is eating food.",),
    ...     ("A man is eating a piece of bread.",),
    ...     ("The girl is carrying a baby.",),
    ...     ("A man is riding a horse.",)
    ... ], ["text"])
    >>> result = pipeline.fit(data).transform(data)
    >>> result.select("ranked_documents").show(truncate=False)
    # Documents will be sorted by relevance with rank information in metadata
    """

    name = "GGUFRankingFinisher"

    inputCols = Param(Params._dummy(),
                     "inputCols",
                     "Name of input annotation columns containing reranked documents",
                     typeConverter=TypeConverters.toListString)

    outputCol = Param(Params._dummy(),
                     "outputCol", 
                     "Name of output annotation column containing ranked documents",
                     typeConverter=TypeConverters.toListString)

    topK = Param(Params._dummy(),
                 "topK",
                 "Maximum number of top documents to return based on relevance score (-1 for no limit)",
                 typeConverter=TypeConverters.toInt)

    minRelevanceScore = Param(Params._dummy(),
                             "minRelevanceScore",
                             "Minimum relevance score threshold for filtering documents",
                             typeConverter=TypeConverters.toFloat)

    minMaxScaling = Param(Params._dummy(),
                         "minMaxScaling",
                         "Whether to apply min-max scaling to normalize relevance scores to 0-1 range",
                         typeConverter=TypeConverters.toBoolean)

    @keyword_only
    def __init__(self):
        super(GGUFRankingFinisher, self).__init__(
            classname="com.johnsnowlabs.nlp.finisher.GGUFRankingFinisher")
        self._setDefault(
            topK=-1,
            minRelevanceScore=float('-inf'),  # Equivalent to Double.MinValue
            minMaxScaling=False,
            outputCol=["ranked_documents"]
        )

    @keyword_only
    def setParams(self):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setInputCols(self, *value):
        """Sets input annotation column names.

        Parameters
        ----------
        value : List[str]
            Input annotation column names containing reranked documents
        """
        if len(value) == 1 and isinstance(value[0], list):
            return self._set(inputCols=value[0])
        else:
            return self._set(inputCols=list(value))

    def getInputCols(self):
        """Gets input annotation column names.

        Returns
        -------
        List[str]
            Input annotation column names
        """
        return self.getOrDefault(self.inputCols)

    def setOutputCol(self, value):
        """Sets output annotation column name.

        Parameters
        ----------
        value : str
            Output annotation column name
        """
        return self._set(outputCol=[value])

    def getOutputCol(self):
        """Gets output annotation column name.

        Returns
        -------
        str
            Output annotation column name
        """
        output_cols = self.getOrDefault(self.outputCol)
        return output_cols[0] if output_cols else "ranked_documents"

    def setTopK(self, value):
        """Sets maximum number of top documents to return.

        Parameters
        ----------
        value : int
            Maximum number of top documents to return (-1 for no limit)
        """
        return self._set(topK=value)

    def getTopK(self):
        """Gets maximum number of top documents to return.

        Returns
        -------
        int
            Maximum number of top documents to return
        """
        return self.getOrDefault(self.topK)

    def setMinRelevanceScore(self, value):
        """Sets minimum relevance score threshold.

        Parameters
        ----------
        value : float
            Minimum relevance score threshold
        """
        return self._set(minRelevanceScore=value)

    def getMinRelevanceScore(self):
        """Gets minimum relevance score threshold.

        Returns
        -------
        float
            Minimum relevance score threshold
        """
        return self.getOrDefault(self.minRelevanceScore)

    def setMinMaxScaling(self, value):
        """Sets whether to apply min-max scaling.

        Parameters
        ----------
        value : bool
            Whether to apply min-max scaling to normalize scores
        """
        return self._set(minMaxScaling=value)

    def getMinMaxScaling(self):
        """Gets whether to apply min-max scaling.

        Returns
        -------
        bool
            Whether min-max scaling is enabled
        """
        return self.getOrDefault(self.minMaxScaling)



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for E5Embeddings."""

from sparknlp.common import *


class E5Embeddings(AnnotatorModel,
                           HasEmbeddingsProperties,
                           HasCaseSensitiveProperties,
                           HasStorageRef,
                           HasBatchedAnnotate,
                           HasMaxSentenceLengthLimit):
    """Sentence embeddings using E5.

    E5, a weakly supervised text embedding model that can generate text embeddings tailored to any task (e.g., classification, retrieval, clustering, text evaluation, etc.)
    Note that this annotator is only supported for Spark Versions 3.4 and up.

    Pretrained models can be loaded with :meth:`.pretrained` of the companion
    object:

    >>> embeddings = E5Embeddings.pretrained() \\
    ...     .setInputCols(["document"]) \\
    ...     .setOutputCol("e5_embeddings")


    The default model is ``"e5_small"``, if no name is provided.

    For available pretrained models please see the
    `Models Hub <https://sparknlp.org/models?q=E5>`__.


    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``DOCUMENT``            ``SENTENCE_EMBEDDINGS``
    ====================== ======================

    Parameters
    ----------
    batchSize
        Size of every batch , by default 8
    dimension
        Number of embedding dimensions, by default 768
    caseSensitive
        Whether to ignore case in tokens for embeddings matching, by default False
    maxSentenceLength
        Max sentence length to process, by default 512
    configProtoBytes
        ConfigProto from tensorflow, serialized into byte array.

    References
    ----------
    `Text Embeddings by Weakly-Supervised Contrastive Pre-training <https://arxiv.org/pdf/2212.03533>`__

    https://github.com/microsoft/unilm/tree/master/e5

    **Paper abstract**

    *This paper presents E5, a family of state-of-the-art text embeddings that transfer
    well to a wide range of tasks. The model is trained in a contrastive manner with
    weak supervision signals from our curated large-scale text pair dataset (called
    CCPairs). E5 can be readily used as a general-purpose embedding model for any
    tasks requiring a single-vector representation of texts such as retrieval, clustering,
    and classification, achieving strong performance in both zero-shot and fine-tuned
    settings. We conduct extensive evaluations on 56 datasets from the BEIR and
    MTEB benchmarks. For zero-shot settings, E5 is the first model that outperforms
    the strong BM25 baseline on the BEIR retrieval benchmark without using any
    labeled data. When fine-tuned, E5 obtains the best results on the MTEB benchmark,
    beating existing embedding models with 40× more parameters.*

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from pyspark.ml import Pipeline
    >>> documentAssembler = DocumentAssembler() \\
    ...     .setInputCol("text") \\
    ...     .setOutputCol("document")
    >>> embeddings = E5Embeddings.pretrained() \\
    ...     .setInputCols(["document"]) \\
    ...     .setOutputCol("e5_embeddings")
    >>> embeddingsFinisher = EmbeddingsFinisher() \\
    ...     .setInputCols(["e5_embeddings"]) \\
    ...     .setOutputCols("finished_embeddings") \\
    ...     .setOutputAsVector(True)
    >>> pipeline = Pipeline().setStages([
    ...     documentAssembler,
    ...     embeddings,
    ...     embeddingsFinisher
    ... ])
    >>> data = spark.createDataFrame([["query: how much protein should a female eat",
    ... "passage: As a general guideline, the CDC's average requirement of protein for women ages 19 to 70 is 46 grams per day." + \
    ... "But, as you can see from this chart, you'll need to increase that if you're expecting or training for a" + \
    ... "marathon. Check out the chart below to see how much protein you should be eating each day.",
    ... ]]).toDF("text")
    >>> result = pipeline.fit(data).transform(data)
    >>> result.selectExpr("explode(finished_embeddings) as result").show(5, 80)
    +--------------------------------------------------------------------------------+
    |                                                                          result|
    +--------------------------------------------------------------------------------+
    |[[8.0190285E-4, -0.005974853, -0.072875895, 0.007944068, 0.026059335, -0.0080...|
    |[[0.050514214, 0.010061974, -0.04340176, -0.020937217, 0.05170225, 0.01157857...|
    +--------------------------------------------------------------------------------+
    """

    name = "E5Embeddings"

    inputAnnotatorTypes = [AnnotatorType.DOCUMENT]

    outputAnnotatorType = AnnotatorType.SENTENCE_EMBEDDINGS
    configProtoBytes = Param(Params._dummy(),
                             "configProtoBytes",
                             "ConfigProto from tensorflow, serialized into byte array. Get with config_proto.SerializeToString()",
                             TypeConverters.toListInt)


    def setConfigProtoBytes(self, b):
        """Sets configProto from tensorflow, serialized into byte array.

        Parameters
        ----------
        b : List[int]
            ConfigProto from tensorflow, serialized into byte array
        """
        return self._set(configProtoBytes=b)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.embeddings.E5Embeddings", java_model=None):
        super(E5Embeddings, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            dimension=768,
            batchSize=8,
            maxSentenceLength=512,
            caseSensitive=False,
        )

    @staticmethod
    def loadSavedModel(folder, spark_session, use_openvino=False):
        """Loads a locally saved model.

        Parameters
        ----------
        folder : str
            Folder of the saved model
        spark_session : pyspark.sql.SparkSession
            The current SparkSession
        use_openvino : bool
            Use OpenVINO backend

        Returns
        -------
        E5Embeddings
            The restored model
        """
        from sparknlp.internal import _E5Loader
        jModel = _E5Loader(folder, spark_session._jsparkSession, use_openvino)._java_obj
        return E5Embeddings(java_model=jModel)

    @staticmethod
    def pretrained(name="e5_small", lang="en", remote_loc=None):
        """Downloads and loads a pretrained model.

        Parameters
        ----------
        name : str, optional
            Name of the pretrained model, by default "e5_small"
        lang : str, optional
            Language of the pretrained model, by default "en"
        remote_loc : str, optional
            Optional remote address of the resource, by default None. Will use
            Spark NLPs repositories otherwise.

        Returns
        -------
        E5Embeddings
            The restored model
        """
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(E5Embeddings, name, lang, remote_loc)
